import argparse
import logging
import os
import time
from typing import Any

import h5py
import matplotlib

from ewoksutils.task_utils import task_inputs
from ewoks import execute_graph

try:
    from ewoksjob.client import submit
except ImportError:
    submit = None

from .utils import (
    NEWFLAT_FILENAME,
    OLDFLAT_FILENAME,
    print_inputs,
)


def streamline_autocalib_workflow_inputs(
    output_dir: str,
    kind: str,
    pyfai_config: str,
    calibrant: str,
    energy: float,
    newflat: str,
    oldflat: str,
    bliss_scan_url: str,
    input_image_url: str,
    max_rings: int | list[int] | None = None,
    isinteractive: bool = False,
) -> list[dict[str, Any]]:
    """Generate the list of inputs for the streamline_autocalib workflow"""
    kind = kind.upper()
    if kind not in ("PDF", "XRD", "SAXS"):
        raise ValueError(f"Unsupported kind: {kind}")

    if max_rings is None:
        if kind == "PDF":
            max_rings = [1, 4, 7, 10, 13, 16, 19, 22, 25, 28, 31, 34, 35]
        elif kind == "XRD":
            max_rings = [1, 4, 7, 10, 13, 16, 19, 20]
        else:
            raise NotImplementedError(f"Unsupported kind: {kind}")

    inputs = []
    inputs += task_inputs(
        task_identifier="FlatFieldFromEnergy",
        inputs={"energy": energy, "newflat": newflat, "oldflat": oldflat},
    )
    inputs += task_inputs(
        task_identifier="PyFaiConfig",
        inputs={
            "filename": pyfai_config,
            "calibrant": calibrant,
        },
    )
    inputs += task_inputs(
        task_identifier="CalibrateSingle",
        inputs={"image": input_image_url, "max_rings": max_rings},
    )
    inputs += task_inputs(  # Both 1D and 2D
        task_identifier="IntegrateSinglePattern",
        inputs={"image": input_image_url},
    )
    inputs += task_inputs(
        task_identifier="SaveNexusPattern1D",
        inputs={
            "url": os.path.join(output_dir, f"result_{kind}.h5"),
            "bliss_scan_url": bliss_scan_url,
        },
    )
    inputs += task_inputs(  # Save 2D
        task_identifier="SaveNexusIntegrated",
        inputs={
            "url": os.path.join(output_dir, f"result_{kind}.h5"),
            "bliss_scan_url": bliss_scan_url,
        },
    )
    inputs += task_inputs(
        task_identifier="DiagnoseCalibrateSingleResults",
        inputs={
            "image": input_image_url,
            "show": isinteractive,
            "filename": os.path.join(output_dir, f"plot_ring_detection_{kind}.svg"),
        },
    )
    inputs += task_inputs(
        task_identifier="DiagnoseIntegrate1D",
        inputs={
            "show": isinteractive,
            "filename": os.path.join(output_dir, f"plot_calibration_{kind}.svg"),
        },
    )
    inputs += task_inputs(
        task_identifier="SavePyFaiPoniFile",
        inputs={
            "output_filename": os.path.join(output_dir, f"calibration_{kind}.poni"),
        },
    )
    inputs += task_inputs(
        task_identifier="SavePyFaiConfig",
        inputs={
            "output_filename": os.path.join(output_dir, f"{kind}.json"),
        },
    )
    return inputs


def create_argument_parser():
    today = time.strftime("%Y%m%d", time.localtime())
    default_session = f"in1176/id31/{today}"
    default_sample = "ESRF_HR_01"
    default_dataset_id = 1

    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "-k",
        "--kind",
        type=str,
        required=True,
        help="Kind of streamline measurment: pdf, xrd, saxs",
    )
    parser.add_argument(
        "--session",
        type=str,
        default=default_session,
        help=f"Session as: <proposal>/<beamline>/<session> (default: {default_session})",
    )
    parser.add_argument(
        "--sample",
        type=str,
        default=default_sample,
        help=f"Sample name (default: {default_sample})",
    )
    parser.add_argument(
        "--dataset",
        type=int,
        default=default_dataset_id,
        help=f"Dataset number (default: {default_dataset_id})",
    )

    calibration_group = parser.add_argument_group("Calibration")
    calibration_group.add_argument(
        "--calibrant",
        type=str,
        required=True,
        help="Name of the calibrant used for calibration image",
    )
    calibration_group.add_argument(
        "--max-rings",
        type=int,
        required=False,
        nargs="+",
        help="Number of rings to use. Use multiple values to refine over an increasing number of rings (default depends on kind argument)",
        default=None,
    )

    runtime_group = parser.add_argument_group("Runtime")
    runtime_group.add_argument(
        "--worker",
        action="store_true",
        help="Execute workflows on ewoks worker instead of current environment",
    )
    runtime_group.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity",
    )
    runtime_group.add_argument(
        "--dry-run",
        action="store_true",
        help="Do not run anything but print which datasets would be processed",
    )

    extra_group = parser.add_argument_group(
        "Extra", "Overrides configuration from session, sample & dataset"
    )
    extra_group.add_argument(
        "-i",
        "--input",
        type=str,
        help="Dataset file to process (HDF5 format)",
        metavar="FILE",
    )
    extra_group.add_argument(
        "-o",
        "--output-dir",
        type=str,
        help="Folder where to store the results",
        metavar="FOLDER",
    )
    extra_group.add_argument(
        "-c",
        "--pyfai-config",
        type=str,
        help="PyFAI config file (.json)",
        metavar="FILE",
    )
    extra_group.add_argument(
        "--flat-dir",
        type=str,
        help="Folder containing flat field files: flats.mat and old_flats.mat",
        metavar="FOLDER",
    )

    return parser


def main():
    logging.basicConfig(level=logging.WARNING)

    parser = create_argument_parser()
    options = parser.parse_args()

    if options.verbose != 0:
        logging.getLogger().setLevel(
            logging.INFO if options.verbose == 1 else logging.DEBUG
        )

    # Use session, sample and dataset to set paths
    session_dir = os.path.join("/data/visitor", options.session)
    input_filename = os.path.join(
        session_dir,
        "RAW_DATA",
        options.sample,
        f"{options.sample}_{options.dataset:04d}",
        f"{options.sample}_{options.dataset:04d}.h5",
    )
    processed_data_dir = os.path.join(session_dir, "PROCESSED_DATA")
    pyfai_config = os.path.join(
        processed_data_dir, "calibration", f"{options.kind.upper()}_ref.json"
    )
    flat_dir = os.path.join(processed_data_dir, "extra_files")
    output_dir = os.path.join(
        processed_data_dir, "calibration", f"{options.kind.upper()}_calibration"
    )

    # Overrides from command line
    if options.input:
        input_filename = os.path.abspath(options.input)
    if options.pyfai_config:
        pyfai_config = options.pyfai_config
    if options.flat_dir:
        flat_dir = os.path.abspath(options.flat_dir)
    if options.output_dir:
        output_dir = os.path.abspath(options.output_dir)

    with h5py.File(input_filename, "r") as h5f:
        energy = h5f["/1.1/instrument/positioners/energy"][()]

    inputs = streamline_autocalib_workflow_inputs(
        output_dir=output_dir,
        kind=options.kind,
        pyfai_config=pyfai_config,
        calibrant=options.calibrant,
        energy=energy,
        newflat=os.path.join(flat_dir, NEWFLAT_FILENAME),
        oldflat=os.path.join(flat_dir, OLDFLAT_FILENAME),
        bliss_scan_url=f"{input_filename}::/1.1",
        input_image_url=f"silx://{input_filename}?path=/1.1/measurement/p3&slice=0",
        max_rings=options.max_rings,
        isinteractive=not options.worker,
    )
    print("Worflow inputs:")
    print_inputs(inputs)

    if options.dry_run:
        return

    if os.path.exists(output_dir):
        raise RuntimeError(f"Output directory already exists: {output_dir}")

    os.makedirs(output_dir)

    print("Execute workflow...")
    workflow = "streamline_autocalib.json"
    workflow_load_options = {"root_module": "ewoksid31.workflows"}
    convert_destination = os.path.join(output_dir, "calibration_workflow.json")
    if options.worker:
        if submit is None:
            raise RuntimeError("Cannot submit workflow to worker: Install ewoksjob")
        future = submit(
            args=(workflow,),
            kwargs={
                "inputs": inputs,
                "convert_destination": convert_destination,
                "load_options": workflow_load_options,
            },
        )
        future.get()
    else:
        # Fix backend set by ewoksxrpd for interactive use
        import ewoksxrpd  # noqa

        matplotlib.use("Qt5Agg")

        _ = execute_graph(
            workflow,
            inputs=inputs,
            convert_destination=convert_destination,
            load_options=workflow_load_options,
        )
    print("Done")


if __name__ == "__main__":
    import sys

    sys.exit(main())



from silx.gui import qt


class ExecutionGroupBox(qt.QGroupBox):

    _LOCAL_HOST = "Local"
    _REMOTE_HOST = "Ewoks worker"

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__("Execution", parent)

        executionModeLayout = qt.QHBoxLayout(self)

        executionModeLabel = qt.QLabel("Host:")
        self._executionModeComboBox = qt.QComboBox()
        self._executionModeComboBox.addItems((self._LOCAL_HOST, self._REMOTE_HOST))
        self._executionModeComboBox.setCurrentText(self._LOCAL_HOST)
        self._executionModeComboBox.setToolTip(
            "Select where the process will be executed: locally or on an Ewoks worker."
        )

        executionModeLayout.addWidget(executionModeLabel)
        executionModeLayout.addWidget(self._executionModeComboBox)
        executionModeLayout.addStretch(1)

    def setLocalExecution(self, local: bool):
        if local:
            self._executionModeComboBox.setCurrentText(self._LOCAL_HOST)
        else:
            self._executionModeComboBox.setCurrentText(self._REMOTE_HOST)

    def isLocalExecution(self) -> bool:
        return self._executionModeComboBox.currentText() == self._LOCAL_HOST



from dataclasses import dataclass
import enum
from typing import Any
import os
from silx.gui import qt
from ewoksutils.task_utils import task_inputs


@dataclass(frozen=True, order=True)
class ScanEntry:
    filename: str
    number: int


class ExportMode(enum.IntEnum):
    DISABLED = 1
    ASCII_SINGLE_DIR = 2
    ASCII_ONE_DIR_PER_SCAN = 3
    HDF5_TOMO = 4


def generateInputs(
    scan: ScanEntry,
    newFlat: str,
    oldFlat: str,
    pyfaiConfigFile: str,
    pyfaiMethod: str,
    monitorName: str | None,
    referenceCounts: int,
    detectorName: str,
    outputDirectory: str,
    sigmaClippingThreshold: float | None,
    exportMode: ExportMode,
    unit: str,
    rotName: str = "",
    yName: str = "",
) -> dict:
    """
    Generate input parameters for the EWOKS workflow, including optional tomo export.
    """
    baseDirName = os.path.splitext(os.path.basename(scan.filename))[0]
    outputUnitSuffix = unit.split("_")[0]
    outputFilePathH5 = os.path.join(outputDirectory, f"{baseDirName}.h5")

    nxProcessName = f"{detectorName}_{outputUnitSuffix}_integrate"
    nxMeasurementName = f"{detectorName}_{outputUnitSuffix}_integrated"

    exportOutputDirectory = os.path.join(outputDirectory, "export")

    asciiBasenameTemplate = (
        f"{baseDirName}_{scan.number:04d}_{detectorName}_{outputUnitSuffix}_%04d.xye"
    )
    if exportMode == ExportMode.ASCII_ONE_DIR_PER_SCAN:
        outputAsciiFileTemplate = os.path.join(
            exportOutputDirectory,
            f"{baseDirName}_{scan.number:04d}",
            asciiBasenameTemplate,
        )
    else:
        outputAsciiFileTemplate = os.path.join(
            exportOutputDirectory, asciiBasenameTemplate
        )

    tomoOutputFilename = os.path.join(
        exportOutputDirectory,
        f"{baseDirName}_{scan.number:04d}_{detectorName}_{outputUnitSuffix}_tomo.h5",
    )

    integrationOptions: dict[str, Any] = {
        "method": pyfaiMethod,
        "unit": unit,
    }
    if sigmaClippingThreshold is not None:
        integrationOptions["extra_options"] = {
            "thres": sigmaClippingThreshold,
            "max_iter": 10,
            "error_model": "azimuthal",
        }
        integrationOptions["integrator_name"] = "sigma_clip_ng"

    inputs = [{"name": "overwrite", "value": True, "all": True}]
    inputs += task_inputs(
        task_identifier="FlatFieldFromEnergy",
        inputs={
            "newflat": newFlat,
            "oldflat": oldFlat,
            "energy": f"{scan.filename}::/{scan.number}.1/instrument/positioners/energy",
            "enabled": detectorName == "p3",
        },
    )
    inputs += task_inputs(
        task_identifier="PyFaiConfig",
        inputs={
            "filename": pyfaiConfigFile,
            "integration_options": integrationOptions,
        },
    )
    inputs += task_inputs(
        task_identifier="IntegrateBlissScan",
        inputs={
            "filename": scan.filename,
            "scan": scan.number,
            "output_filename": outputFilePathH5,
            "monitor_name": monitorName,
            "reference": referenceCounts,
            "maximum_persistent_workers": 2,
            "retry_timeout": 3600,
            "detector_name": detectorName,
            "nxprocess_name": nxProcessName,
            "nxmeasurement_name": nxMeasurementName,
        },
    )
    inputs += task_inputs(
        task_identifier="SaveNexusPatternsAsAscii",
        inputs={
            "enabled": exportMode
            in (ExportMode.ASCII_SINGLE_DIR, ExportMode.ASCII_ONE_DIR_PER_SCAN),
            "output_filename_template": outputAsciiFileTemplate,
        },
    )
    inputs += task_inputs(
        task_identifier="SaveNexusPatternsAsId31TomoHdf5",
        inputs={
            "enabled": exportMode == ExportMode.HDF5_TOMO,
            "scan_entry_url": f"{scan.filename}::/{scan.number}.1",
            "rot_name": rotName,
            "y_name": yName,
            "output_filename": tomoOutputFilename,
        },
    )

    return {
        "inputs": inputs,
        "convert_destination": os.path.join(
            outputDirectory,
            f"{baseDirName}_{scan.number}_{detectorName}_{outputUnitSuffix}.json",
        ),
    }


def extractScanNumber(h5path: str) -> int:
    """
    Extracts the scan number from the h5path from a selected node.

    Example: '/2.1/measurement/p3" -> returns 2'
    """
    parts = h5path.split("/")
    if len(parts) > 1:
        scanNumberPart = parts[1].split(".")[0]
        try:
            return int(scanNumberPart)
        except ValueError:
            pass
    return -1


def generateUniquePath(basePath: str) -> str:
    """
    Generates a unique path by adding an incremantal suffix if necessary.

    Args:
        basePath: Base path (file or directory)

    Returns:
        unique path
    """
    isFile = os.path.isfile(basePath)
    dirName = os.path.dirname(basePath)
    baseName, ext = (
        os.path.splitext(os.path.basename(basePath)) if isFile else (basePath, "")
    )

    parts = baseName.rsplit("_", 1)
    if len(parts) == 2 and parts[1].isdigit():
        baseName, counter = parts[0], int(parts[1]) + 1
    else:
        counter = 1

    generatedName = os.path.join(dirName, f"{baseName}_{counter}{ext}")
    while os.path.exists(generatedName):
        counter += 1
        generatedName = os.path.join(dirName, f"{baseName}_{counter}{ext}")
    return generatedName


class FilenameCompleterLineEdit(qt.QLineEdit):
    """
    Heritage from QLineEdit widget that provides autocompletion for file paths.

    This widget uses a QFileSystemModel to suggest file and directory paths,
    starting from the root of the filesystem.
    """

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent=parent, **kwargs)

        completer = qt.QCompleter()
        model = qt.QFileSystemModel(completer)
        model.setOption(qt.QFileSystemModel.Option.DontWatchForChanges, True)
        model.setRootPath("/")

        completer.setModel(model)
        completer.setCompletionRole(qt.QFileSystemModel.Roles.FileNameRole)
        self.setCompleter(completer)


def get_scan_url(workflow_parameters):
    datasetFilename = None
    scanNumber = None
    for item in workflow_parameters["inputs"]:
        if item.get("task_identifier") != "IntegrateBlissScan":
            continue
        if item["name"] == "filename":
            datasetFilename = item["value"]
        if item["name"] == "scan":
            scanNumber = item["value"]

    return datasetFilename, scanNumber



from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.gui.parameterform import ParameterForm
from ewoksid31.tasks.flatfield import FlatFieldFromEnergy


class OWFlatFieldFromEnergy(  # type: ignore[call-arg]
    OWEwoksWidgetOneThread,
    ewokstaskclass=FlatFieldFromEnergy,
):
    name = "FlatFieldFromEnergy"
    description = "Interpolate an energy-stack of flat-field images"
    icon = "icons/widget.png"
    want_main_area = False

    def __init__(self):
        super().__init__()
        self._init_control_area()

        self._parameter_form = ParameterForm(parent=self.controlArea)
        self._parameter_form.addParameter(
            "newflat",
            label="New flat filename",
            value_for_type="",
            select="file",
            value_change_callback=self._inputs_changed,
        )
        self._parameter_form.addParameter(
            "oldflat",
            label="Old flat filename",
            value_for_type="",
            select="file",
            value_change_callback=self._inputs_changed,
        )
        self._parameter_form.addParameter(
            "energy",
            label="X-Ray energy in KeV",
            value_for_type=0.0,
            value_change_callback=self._inputs_changed,
        )
        self._update_parameter_values()

        self.controlArea.layout().addStretch(1)

    def _inputs_changed(self):
        new_values = self._parameter_form.get_parameter_values()
        self.update_default_inputs(**new_values)

    def _update_parameter_values(self):
        initial_values = self.get_default_input_values()
        self._parameter_form.set_parameter_values(initial_values)



from silx.gui import qt

from .workflow_executor import JobItem, WorkflowExecutor


class JobsProgressBar(qt.QProgressBar):
    def __init__(
        self, executor: WorkflowExecutor, parent: qt.QWidget | None = None
    ) -> None:
        super().__init__(parent)
        self.setFormat("%v / %m")
        self.setMaximumWidth(self.sizeHint().width() * 2)
        self.setEnabled(False)

        executor.finished.connect(self._finished)
        executor.jobSubmitted.connect(self._jobSubmitted)
        executor.jobChanged.connect(self._jobChanged)

    def _finished(self) -> None:
        self.setEnabled(False)

    def _jobSubmitted(self) -> None:
        if self.isEnabled():
            self.setMaximum(self.maximum() + 1)
            return

        self.setRange(0, 1)
        self.setValue(0)
        self.setEnabled(True)

    def _jobChanged(self, jobItem: JobItem) -> None:
        future = jobItem.getFuture()
        if future.cancelled():
            self.setMaximum(max(0, self.maximum() - 1))
        elif future.done():
            self.setValue(self.value() + 1)



FILE_EXTENSIONS = ".h5", ".hdf5", ".hdf", ".nx", ".nxs", ".nx5", ".nexus"

MONITOR_NAMES: tuple[str, str, str, str] = (
    "mondio",
    "scaled_detdio",
    "scaled_mondio",
    "srcur",
)

DETECTOR_NAMES: tuple[str, str, str] = (
    "p3",
    "de",
    "perkin",
)
INTEGRATION_METHODS: tuple[str, ...] = (
    "no_csr_cython",
    "bbox_csr_cython",
    "full_csr_cython",
    "no_csr_ocl_gpu",
    "bbox_csr_ocl_gpu",
    "full_csr_ocl_gpu",
    "no_histogram_cython",
    "bbox_histogram_cython",
    "full_histogram_ocl_gpu",
)



from typing import Literal
import functools
import logging
import os
import qtawesome
from silx.gui import icons, qt

from ewoksjob.client.futures import CancelledError, FutureInterface, TimeoutError
from .workflow_executor import JobItem, WorkflowExecutor
from .utils import get_scan_url


_logger = logging.getLogger(__name__)


_FutureStateStr = Literal["pending", "running", "succeeded", "failed", "cancelled"]


class _RunningIconItems:
    """Set an animated 'running' icon to all contained QListWidgetItems"""

    def __init__(self) -> None:
        self._waitIcon = icons.getWaitIcon()
        self._items: list[qt.QListWidgetItem] = list()

    def add(self, item: qt.QListWidgetItem) -> None:
        if not self._items:
            self._waitIcon.register(self)
            self._waitIcon.iconChanged.connect(self._iconChanged)

        if item not in self._items:
            item.setIcon(self._waitIcon.currentIcon())
            self._items.append(item)

    def discard(self, item: qt.QListWidgetItem) -> None:
        try:
            self._items.remove(item)
        except ValueError:
            return

        if len(self._items) == 0:
            self._waitIcon.unregister(self)
            self._waitIcon.iconChanged.disconnect(self._iconChanged)

    def _iconChanged(self, icon: qt.QIcon) -> None:
        for item in self._items:
            item.setIcon(icon)


class JobsListWidget(qt.QWidget):

    def __init__(
        self, executor: WorkflowExecutor, parent: qt.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self._listWidget = qt.QListWidget()
        self._listWidget.setSelectionMode(
            qt.QAbstractItemView.SelectionMode.MultiSelection
        )
        self._listWidget.setDragDropMode(qt.QAbstractItemView.DragDropMode.NoDragDrop)
        self._listWidget.itemSelectionChanged.connect(self._itemSelectionChanged)

        toolbar = qt.QToolBar()
        toolbar.setToolButtonStyle(qt.Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        toolbar.setContentsMargins(0, 0, 0, 0)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Preferred)
        toolbar.addWidget(spacer)

        self._cancelAction = qt.QAction(self)
        self._cancelAction.setIcon(qtawesome.icon("fa6s.ban", scale_factor=0.75))
        self._cancelAction.setText("Cancel Selected")
        self._cancelAction.setEnabled(False)
        self._cancelAction.triggered.connect(self._cancelSelected)
        toolbar.addAction(self._cancelAction)

        clearAction = qt.QAction(self)
        clearAction.setIcon(qtawesome.icon("fa6s.broom", scale_factor=0.75))
        clearAction.setText("Clear Completed")
        clearAction.setToolTip("Remove all completed/failed jobs")
        clearAction.triggered.connect(self.removeFinishedItems)
        toolbar.addAction(clearAction)

        layout = qt.QVBoxLayout(self)
        layout.setSpacing(0)
        layout.addWidget(self._listWidget, stretch=1)
        layout.addWidget(toolbar)

        self._runningIconItems = _RunningIconItems()

        executor.jobSubmitted.connect(self._jobSubmitted)
        executor.jobChanged.connect(self._jobChanged)

    def removeFinishedItems(self) -> None:
        # Look for items to remove in reversed order
        # to avoid changing the row of the following items to be checked for removal
        for row in reversed(range(self._listWidget.count())):
            listItem = self._listWidget.item(row)
            if listItem is None:
                continue
            jobItem: JobItem = listItem.data(qt.Qt.UserRole)
            if jobItem.getFuture().done():
                self._listWidget.takeItem(row)

    def _cancelSelected(self) -> None:
        for listItem in self._listWidget.selectedItems():
            jobItem = listItem.data(qt.Qt.UserRole)
            if jobItem is None:
                continue
            jobItem.getFuture().cancel()

    def _itemSelectionChanged(self) -> None:
        self._cancelAction.setEnabled(len(self._listWidget.selectedItems()) > 0)

    def _updateListItem(self, listItem: qt.QListWidgetItem, jobItem: JobItem) -> None:
        state = self._getFutureState(jobItem.getFuture())

        filename, number = get_scan_url(jobItem.getArguments())
        timestamp = jobItem.getTimestamp().time().isoformat("minutes")
        listItem.setText(f"{timestamp} {os.path.basename(filename)}::/{number}.1")
        listItem.setToolTip(
            f"<b>{state.capitalize()}</b><br><b>Scan:</b> {filename}::/{number}.1"
        )

        if state in ("failed", "cancelled"):
            listItem.setForeground(qt.Qt.red)

        if state == "pending":
            listItem.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
        else:
            # Unselect manually since removing the ItemIsSelectable flag does not unselect the item
            if listItem.isSelected():
                listItem.setSelected(False)
            listItem.setFlags(qt.Qt.ItemIsEnabled)

        if state == "running":
            self._runningIconItems.add(listItem)
            return

        self._runningIconItems.discard(listItem)
        listItem.setIcon(self._getStateIcon(state))

    @staticmethod
    @functools.cache
    def _getStateIcon(state: _FutureStateStr) -> qt.QIcon:
        iconNames: dict[_FutureStateStr, str | None] = {
            "pending": "fa6.hourglass-half",
            "succeeded": "fa6s.check",
            "cancelled": "fa6s.ban",
            "failed": "fa6s.exclamation",
            "running": None,
        }
        color = "red" if state in ("cancelled", "failed") else None
        return qtawesome.icon(iconNames[state], color=color, scale_factor=0.75)

    @staticmethod
    def _getFutureState(future: FutureInterface) -> _FutureStateStr:
        if future.running():
            return "running"

        if not future.done():
            return "pending"

        try:
            future.result(0.001)
        except TimeoutError:
            _logger.error("Cannot get workflow results")
            return "failed"
        except CancelledError:
            return "cancelled"
        except Exception:
            return "failed"
        else:
            return "succeeded"

    def _jobSubmitted(self, jobItem: JobItem) -> None:
        listItem = qt.QListWidgetItem()
        listItem.setData(qt.Qt.UserRole, jobItem)
        self._updateListItem(listItem, jobItem)
        self._listWidget.addItem(listItem)
        self._listWidget.scrollToBottom()

    def _jobChanged(self, jobItem: JobItem) -> None:
        listItem = self._findListItem(jobItem)
        if listItem is None:
            return
        self._updateListItem(listItem, jobItem)

    def _findListItem(self, jobItem: JobItem) -> qt.QListWidgetItem | None:
        for row in range(self._listWidget.count()):
            listItem = self._listWidget.item(row)
            if listItem is None:
                continue
            if listItem.data(qt.Qt.UserRole) is jobItem:
                return listItem
        return None



import h5py
import numpy

from .resources import resource_filename
from ..tasks.flatfield import FlatFieldFromEnergy


def test_flatfield():
    inputs = {
        "newflat": resource_filename("flats.mat"),
        "oldflat": resource_filename("flats_old.mat"),
        "energy": 90.0,
    }
    task = FlatFieldFromEnergy(inputs=inputs)
    task.execute()
    # Flat-field correction:
    #  Icor = I * 0.8115191925457376 = I / flatfield
    numpy.testing.assert_allclose(1 / task.outputs.flatfield[0, 0], 0.8115191925457376)


def test_flatfield_energy_url(tmp_path):
    filepath = tmp_path / "flatfield.h5"
    datapath = "/1.1/instrument/energy"
    with h5py.File(filepath, "w") as h5f:
        h5f[datapath] = 90

    inputs = {
        "newflat": resource_filename("flats.mat"),
        "oldflat": resource_filename("flats_old.mat"),
        "energy": f"{str(filepath)}::{datapath}",
    }

    task = FlatFieldFromEnergy(inputs=inputs)
    task.execute()
    # Flat-field correction:
    #  Icor = I * 0.8115191925457376 = I / flatfield
    numpy.testing.assert_allclose(1 / task.outputs.flatfield[0, 0], 0.8115191925457376)


def test_flatfield_disabled():
    inputs = {
        "newflat": resource_filename("flats.mat"),
        "oldflat": resource_filename("flats_old.mat"),
        "energy": 90.0,
        "enabled": False,
    }
    task = FlatFieldFromEnergy(inputs=inputs)
    task.execute()
    assert task.outputs.flatfield is None



import sys
import os
import signal
import logging
import ctypes
import argparse

# Environment variable HDF5_USE_FILE_LOCKING must be set *before* importing HDF5 libraries
# This prevents file locking issues, especially in shared filesystems (e.g., NFS).
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

import hdf5plugin  # noqa: F401, E402 - Imported but unused

from silx.gui import qt  # noqa: E402
from .mainwindow import MainWindow  # noqa: E402
from ..utils import FLATFIELD_DEFAULT_DIR  # noqa: E402
from .constants import FILE_EXTENSIONS  # noqa: E402

_logger = logging.getLogger(__name__)


def create_argument_parser():
    parser = argparse.ArgumentParser(
        description="",
    )
    parser.add_argument(
        "-f",
        "--fresh",
        action="store_true",
        help="Start without loading previous user preferences",
    )
    parser.add_argument(
        "default_dir",
        nargs="?",
        type=str,
        help="Default directory for raw/config/output dialogs (ex: /data/visitor/in1176/id31/20250412)",
        default=None,
        metavar="FOLDER",
    )

    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=False,
        help="Dataset file to process (HDF5 format)",
        default="",
        metavar="FILE",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        type=str,
        required=False,
        help="Folder where to store the results",
        default="",
        metavar="FOLDER",
    )
    parser.add_argument(
        "-c",
        "--pyfai-config",
        type=str,
        default=None,
        help="PyFAI config file (.json)",
        metavar="FILE",
    )
    parser.add_argument(
        "--flat-dir",
        type=str,
        required=False,
        help=f"Folder containing flat-field files: flats.mat and old_flats.mat (default: {FLATFIELD_DEFAULT_DIR})",
        default=FLATFIELD_DEFAULT_DIR,
        metavar="FOLDER",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity level (-v: INFO, -vv:DEBUG)",
    )
    return parser


def main() -> int:
    logging.basicConfig(level=logging.WARNING)

    parser = create_argument_parser()
    args = parser.parse_args()

    logging.captureWarnings(True)

    if args.verbose == 0:
        logging.getLogger().setLevel(logging.ERROR)
        _logger.setLevel(logging.INFO)

    elif args.verbose == 1:
        logging.getLogger().setLevel(logging.WARNING)
        _logger.setLevel(logging.INFO)

    else:
        logging.getLogger().setLevel(logging.DEBUG)

    if sys.platform == "win32":
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
            "ESRF.id31pyfaiewoks"
        )

    app = qt.QApplication([])
    app.setOrganizationName("ESRF")
    app.setOrganizationDomain("esrf.fr")
    app.setApplicationName("id31pyfaiewoks")

    app.setWindowIcon(
        qt.QIcon(os.path.join(os.path.dirname(__file__), "integrate.svg"))
    )

    window = MainWindow()
    window.setAttribute(qt.Qt.WA_DeleteOnClose)

    window.setFlatFieldDirName(args.flat_dir)

    if not args.fresh:
        _logger.debug("Loading settings from previous session.")
        window.loadSettings()

    if args.default_dir:
        default_dir = os.path.abspath(args.default_dir)
        if os.path.isdir(default_dir):
            window.setDefaultDirectory(default_dir)
            window.setConfigFilePath("")
            window._outputWidget.setDefaultDirectory(default_dir)
            window._outputWidget.setOutputDirName(default_dir)
            _logger.info(f"Set default directory: {default_dir}")
        else:
            _logger.warning(f"Invalid default directory provided: {default_dir}")

    if args.input:
        raw_data = os.path.abspath(args.input)
        if os.path.isfile(raw_data) and raw_data.endswith(FILE_EXTENSIONS):
            window.addRawDataFile(raw_data)
        else:
            _logger.error(f"Invalid raw data file path or format: {raw_data}")

    if args.output_dir:
        window._outputWidget.setOutputDirName(args.output_dir)

    if args.pyfai_config:
        config_file = os.path.abspath(args.pyfai_config)
        if os.path.isfile(config_file):
            window.setConfigFilePath(config_file)
        else:
            _logger.error(f"Invalid config file path: {config_file}")

    window.show()

    result = app.exec_()
    if window.isWorkflowRunning():
        # Running local threads cannot be cancelled, kill the application
        os.kill(os.getpid(), signal.SIGKILL)
    return result



import h5py
import numpy as np
import pytest

from ewoksid31.tasks.tomo import SaveNexusPatternsAsId31TomoHdf5


def test_SaveNexusPatternsAsId31TomoHdf5(
    tmp_path,
    tomo_scans,
    integrated_scan,
):

    output_filename = str(tmp_path / "result.h5")
    inputs = {
        "scan_entry_url": f"{tomo_scans}::/2.1",
        "rot_name": "nth",
        "y_name": "ny",
        "nxdata_url": f"{integrated_scan}::/integrated",
        "output_filename": output_filename,
    }

    task = SaveNexusPatternsAsId31TomoHdf5(inputs=inputs)
    task.execute()
    assert task.outputs.filename == output_filename

    with h5py.File(integrated_scan, "r") as integrated_file:
        q = integrated_file["integrated/q"][()]
        intensity = integrated_file["integrated/intensity"][()]

    with h5py.File(tomo_scans, "r") as raw_data_file:
        nth = raw_data_file["2.1/measurement/nth"][()]
        ny = raw_data_file["2.1/measurement/ny"][()]

    with h5py.File(output_filename) as h5f:
        assert np.array_equal(h5f["XRD"][()], intensity)
        assert np.array_equal(h5f["q"][()], q)
        assert h5f["q"].attrs["units"] == "nm^-1"
        assert np.array_equal(h5f["th"][()], nth)
        assert h5f["th"].attrs["units"] == "deg"
        assert np.array_equal(h5f["y"][()], ny)
        assert h5f["y"].attrs["units"] == "mm"


def test_overwrite_SaveNexusPatternsAsId31TomoHdf5(
    tmp_path,
    tomo_scans,
    integrated_scan,
):

    output_filename = str(tmp_path / "result.h5")

    # Create output file to simulate conflict
    with h5py.File(output_filename, "w") as h5f:
        h5f.create_group("dummy")

    inputs = {
        "scan_entry_url": f"{tomo_scans}::/2.1",
        "rot_name": "nth",
        "y_name": "ny",
        "nxdata_url": f"{integrated_scan}::/integrated",
        "output_filename": output_filename,
    }

    # Check that it fails without overwrite
    task_without_overwrite = SaveNexusPatternsAsId31TomoHdf5(inputs=inputs)
    with pytest.raises(RuntimeError) as exc:
        task_without_overwrite.execute()
    original_exc = exc.value.__cause__
    assert isinstance(original_exc, FileExistsError)
    assert (
        str(original_exc)
        == f"{output_filename} already exists. Use overwrite=True to replace it."
    )

    # Check that it pass with overwrite=True
    task_with_overwrite = SaveNexusPatternsAsId31TomoHdf5(
        inputs={
            **inputs,
            "overwrite": True,
        }
    )
    task_with_overwrite.execute()
    assert task_with_overwrite.outputs.filename == output_filename



import os.path
import h5py

import silx.gui.hdf5
from silx.gui.hdf5._utils import H5Node
from silx.gui import qt


class Hdf5TreeView(silx.gui.hdf5.Hdf5TreeView):

    h5NodeActivated = qt.Signal(H5Node)
    """Signal emitted when a H5Node is activated for viewing"""

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__(parent)
        self.setSelectionMode(qt.QAbstractItemView.SelectionMode.MultiSelection)
        self.setDragDropMode(qt.QAbstractItemView.DragDropMode.NoDragDrop)
        self.addContextMenuCallback(self.__customContextMenu)
        self.activated.connect(self.__activated)

    def selectionCommand(
        self, index: qt.QModelIndex, event: qt.QEvent | None
    ) -> qt.QItemSelectionModel.SelectionFlag:
        """Enable item selection only with left mouse button"""
        if (
            event is not None
            and event.type() == qt.QEvent.Type.MouseButtonPress
            and event.button() != qt.Qt.LeftButton
        ):
            return qt.QItemSelectionModel.SelectionFlag.NoUpdate
        return super().selectionCommand(index, event)

    def __activated(self, index: qt.QModelIndex):
        if not index.isValid():
            return
        firstIndex = index.siblingAtColumn(0)
        if not firstIndex.isValid():
            return
        item = self.model().data(firstIndex, silx.gui.hdf5.Hdf5TreeModel.H5PY_ITEM_ROLE)
        if item is None:
            return
        self.h5NodeActivated.emit(H5Node(item))

    def __customContextMenu(self, event: silx.gui.hdf5.Hdf5ContextMenuEvent):
        """
        Populate the custom context menu for the HDF5 tree viewer.

        Adds additional options to remove or reload HDF5 files in the tree view when the
        context menu is triggered.
        """
        menu = event.menu()

        if not menu.isEmpty():
            menu.addSeparator()

        h5node = event.hoveredObject()
        action = qt.QAction("View", event.source())
        action.triggered.connect(lambda: self.h5NodeActivated.emit(h5node))
        menu.addAction(action)

        if h5node.ntype is h5py.File:
            filename = os.path.basename(h5node.local_filename)
            action = qt.QAction(f"Remove {filename}", event.source())
            action.triggered.connect(
                lambda: self.findHdf5TreeModel().removeH5pyObject(h5node.h5py_object)
            )
            menu.addAction(action)
            action = qt.QAction(f"Reload {filename}", event.source())
            action.triggered.connect(
                lambda: self.findHdf5TreeModel().synchronizeH5pyObject(
                    h5node.h5py_object
                )
            )
            menu.addAction(action)



import os
import logging

from silx.gui import qt, icons

from .utils import ExportMode, FilenameCompleterLineEdit

_logger = logging.getLogger(__name__)


class OutputWidget(qt.QGroupBox):
    def __init__(self):
        super().__init__("Output")

        self._outputLineEdit = FilenameCompleterLineEdit()
        self._outputLineEdit.setPlaceholderText("/path/to/output/directory")

        self._defaultDirectory = os.getcwd()

        outputDirButton = qt.QPushButton("Select...")
        outputDirButton.setToolTip("Define a new output directory.")
        outputDirButton.setIcon(icons.getQIcon("folder"))
        outputDirButton.clicked.connect(self._outputDirButtonClicked)

        exportDisableRadioButton = qt.QRadioButton("Disabled")
        exportAsciiRadioButton = qt.QRadioButton("ASCII (.xye)")
        self._exportAsciiScanFoldersCheckbox = qt.QCheckBox("Grouped per scan")

        exportHdf5TomoRadioButton = qt.QRadioButton("HDF5 Tomo (.h5):")
        self._exportButtonGroup = qt.QButtonGroup()

        self._exportButtonGroup.addButton(exportDisableRadioButton, 1)
        self._exportButtonGroup.addButton(exportAsciiRadioButton, 2)
        self._exportButtonGroup.addButton(exportHdf5TomoRadioButton, 3)

        exportDisableRadioButton.setChecked(True)
        self._exportAsciiScanFoldersCheckbox.setEnabled(False)
        exportAsciiRadioButton.toggled.connect(
            self._exportAsciiScanFoldersCheckbox.setEnabled
        )

        self._rotNameLineEdit = qt.QLineEdit()
        self._rotNameLineEdit.setPlaceholderText("rot_motor_name")
        self._rotNameLineEdit.setToolTip(
            "Specify the Rotation angle dataset (e.g., nth)."
        )

        self._yNameLineEdit = qt.QLineEdit()
        self._yNameLineEdit.setPlaceholderText("y_motor_name")
        self._yNameLineEdit.setToolTip(
            "Specify the radial position dataset (e.g., ny)."
        )

        self._tomoMotorWidget = qt.QWidget(self)
        self._tomoMotorWidget.setEnabled(False)
        exportHdf5TomoRadioButton.toggled.connect(self._tomoMotorWidget.setEnabled)

        tomoMotorLayout = qt.QHBoxLayout(self._tomoMotorWidget)
        tomoMotorLayout.setContentsMargins(0, 0, 0, 0)
        tomoMotorLayout.addWidget(qt.QLabel("Rot name:"))
        tomoMotorLayout.addWidget(self._rotNameLineEdit)
        tomoMotorLayout.addWidget(qt.QLabel("Y name:"))
        tomoMotorLayout.addWidget(self._yNameLineEdit)
        tomoMotorLayout.addStretch(1)

        gridLayout = qt.QGridLayout(self)
        gridLayout.setColumnStretch(2, 1)

        gridLayout.addWidget(qt.QLabel("Directory:"), 0, 0, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(self._outputLineEdit, 0, 1, 1, 2)
        gridLayout.addWidget(outputDirButton, 0, 3, 1, 1, qt.Qt.AlignRight)

        gridLayout.addWidget(qt.QLabel("Export:"), 1, 0, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(exportDisableRadioButton, 1, 1, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(exportAsciiRadioButton, 2, 1, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(
            self._exportAsciiScanFoldersCheckbox, 2, 2, 1, 1, qt.Qt.AlignLeft
        )
        gridLayout.addWidget(exportHdf5TomoRadioButton, 3, 1, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(self._tomoMotorWidget, 3, 2, 1, 2, qt.Qt.AlignLeft)

    def getDefaultDirectory(self) -> str:
        """
        Return the default directory used when no output directory is set.
        """
        return self._defaultDirectory

    def setDefaultDirectory(self, path: str) -> None:
        """
        Set the default directory used when no output directory is set.
        """
        self._defaultDirectory = path

    def getOutputDirName(self) -> str:
        """
        Returns the current output directory path from the line edit.
        """
        return self._outputLineEdit.text().strip()

    def setOutputDirName(self, path: str) -> None:
        """
        Update the output directory path in the line edit.
        """
        self._outputLineEdit.setText(os.path.abspath(path))

    def _outputDirButtonClicked(self) -> None:
        """
        Choose an output directory using a dialog and set it in the line edit.
        """
        currentPath = self.getOutputDirName()
        # if not currentPath or not os.path.exists(currentPath):
        #     currentPath = self._defaultDirectory()
        if not currentPath or not os.path.exists(currentPath):
            currentPath = self.getDefaultDirectory()

        newPath = qt.QFileDialog.getExistingDirectory(
            self,
            "Choose output directory",
            currentPath,
        )
        if newPath:
            self.setOutputDirName(newPath)

    def getExportMode(self) -> ExportMode:
        checkedId = self._exportButtonGroup.checkedId()
        if checkedId == 1:
            return ExportMode.DISABLED
        elif checkedId == 2:
            if self._exportAsciiScanFoldersCheckbox.isChecked():
                return ExportMode.ASCII_ONE_DIR_PER_SCAN
            else:
                return ExportMode.ASCII_SINGLE_DIR
        elif checkedId == 3:
            return ExportMode.HDF5_TOMO
        else:
            _logger.warning(f"Unknown buttonId: {checkedId}")
            return ExportMode.DISABLED

    def setExportMode(self, mode: ExportMode):
        if mode == ExportMode.DISABLED:
            self._exportButtonGroup.button(1).setChecked(True)
        elif mode == ExportMode.ASCII_SINGLE_DIR:
            self._exportButtonGroup.button(2).setChecked(True)
        elif mode == ExportMode.ASCII_ONE_DIR_PER_SCAN:
            self._exportButtonGroup.button(2).setChecked(True)
        elif mode == ExportMode.HDF5_TOMO:
            self._exportButtonGroup.button(3).setChecked(True)
        self._exportAsciiScanFoldersCheckbox.setChecked(
            mode == ExportMode.ASCII_ONE_DIR_PER_SCAN
        )

    def getTomoRotName(self) -> str:
        return self._rotNameLineEdit.text().strip()

    def setTomoRotName(self, value: str):
        self._rotNameLineEdit.setText(value)

    def getTomoYName(self) -> str:
        return self._yNameLineEdit.text().strip()

    def setTomoYName(self, value: str):
        self._yNameLineEdit.setText(value)



import logging
import numpy
from pydantic import Field
from scipy.io import loadmat
from ewokscore import Task
from ewokscore.model import BaseInputModel
import ewoksdata.data.bliss


_logger = logging.getLogger(__name__)


class Inputs(BaseInputModel):
    newflat: str = Field(
        description="Filename of ID31 P3 detector flatfield",
        examples=["/data/id31/inhouse/P3/flats.mat"],
    )
    oldflat: str = Field(
        description="Filename of ID31 P3 detector 'old' flatfield",
        examples=["/data/id31/inhouse/P3/flats_old.mat"],
    )
    energy: float | str = Field(
        description="X-ray energy in KeV or URL of hdf5 dataset storing its value",
        examples=[75.0, "/path/to/file.f5::/2.1/instrument/positioners/energy"],
    )
    enabled: bool = True


class FlatFieldFromEnergy(Task, input_model=Inputs, output_names=["flatfield"]):  # type: ignore[call-arg]
    """Interpolate an energy-stack of flat-field images.

    The resulting flat-field image can be used as follows
    to correct diffraction patterns for flat field:

    .. code::

        Icor = I / flatfield

    Output:

    - flatfield: 2D numpy.ndarray
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())

        if not inputs.enabled:
            _logger.info(
                f"Task {self.__class__.__qualname__} is disabled: No flatfield"
            )
            self.outputs.flatfield = None
            return

        if isinstance(inputs.energy, str):
            energy = ewoksdata.data.bliss.get_data(inputs.energy)
        else:
            energy = inputs.energy

        new_flat = _interpolated_flatfield(inputs.newflat, "E", "F", energy)
        old_flat = _interpolated_flatfield(inputs.oldflat, "Eold", "Fold", energy)
        flatfield = old_flat / new_flat
        flatfield[~numpy.isfinite(flatfield)] = 1
        flatfield[flatfield < 0] = 1
        self.outputs.flatfield = flatfield


def _interpolated_flatfield(
    matlab_filename: str, energy_key: str, flatfield_key: str, energy: float
) -> numpy.ndarray:
    """
    :param matlab_filename: matlab file from ID31 that contains an energy-stack of images
    :param energy_key:
    :param flatfield_key:
    :param energy:
    :return: interpolated image (nrow, ncol)
    """
    m = loadmat(matlab_filename)
    flatfields = m[flatfield_key]
    energies = numpy.squeeze(m[energy_key]).astype(float)
    return _interpolate_flatfields(energies, flatfields, energy)


def _interpolate_flatfields(
    energies: numpy.ndarray, flatfields: numpy.ndarray, energy: float
) -> numpy.ndarray:
    """
    :param energies: array of energies (nenergies,)
    :param flatfields: stack of images (nrow, ncol, nenergies)
    :param energy:
    :return: interpolated image (nrow, ncol)
    """
    j = numpy.argsort(abs(energies - energy))[0:2]
    j.sort()
    j_before, j_after = j
    e_before = energies[j_before]
    e_after = energies[j_after]
    w_before = (e_after - energy) / (e_after - e_before)
    w_after = (energy - e_before) / (e_after - e_before)
    f_before = flatfields[..., j_before]
    f_after = flatfields[..., j_after]
    return w_before * f_before + w_after * f_after



from datetime import datetime
from pathlib import Path

import h5py
import numpy
import pytest

from silx.io.dictdump import dicttonx

DETECTOR_SHAPE = (1679, 1475)


@pytest.fixture()
def tomo_scans(tmpdir, n_frames=10):
    scan_dir = Path(tmpdir / "RAW_DATA")
    scan_dir.mkdir()
    scan_file = scan_dir / "scan.h5"
    with h5py.File(scan_file, "w") as h5f:
        scan_group = h5f.create_group("/2.1")
        dicttonx(
            {
                "end_time": str(datetime.now()),
                "instrument": {
                    "nth": {
                        "data": numpy.linspace(90, 45, n_frames),
                        "data@units": "deg",
                    },
                    "ny": {"data": numpy.linspace(0, 2, n_frames), "data@units": "mm"},
                    "p3": {"data": numpy.zeros((n_frames, *DETECTOR_SHAPE))},
                    "positioners": {"energy": 75.0},
                },
                "measurement": {
                    ">nth": "../instrument/nth/data",
                    ">ny": "../instrument/ny/data",
                    ">p3": "../instrument/p3/data",
                    "scaled_mondio": numpy.ones((n_frames,)),
                },
            },
            scan_group,
        )
        scan_group = h5f.create_group("/3.1")
        dicttonx(
            {
                "end_time": str(datetime.now()),
                "instrument": {
                    "nth": {
                        "data": numpy.linspace(90, 45, n_frames),
                        "data@units": "deg",
                    },
                    "ny": {"data": numpy.linspace(0, 2, n_frames), "data@units": "mm"},
                    "perkin": {"data": numpy.zeros((n_frames, *DETECTOR_SHAPE))},
                    "positioners": {"energy": 75.0},
                },
                "measurement": {
                    ">nth": "../instrument/nth/data",
                    ">ny": "../instrument/ny/data",
                    ">perkin": "../instrument/perkin/data",
                    "scaled_mondio": numpy.ones((n_frames,)),
                },
            },
            scan_group,
        )

    return scan_file


@pytest.fixture
def integrated_scan(tmp_path) -> Path:
    intensities = numpy.ones((6, 100))
    q = numpy.arange(100)

    # Create integrated.h5
    integrated_patterns_file = tmp_path / "integrated.h5"
    with h5py.File(integrated_patterns_file, "w") as h5f:
        dicttonx(
            {
                "integrated": {
                    "@NX_class": "NXdata",
                    "@signal": "intensity",
                    "@axes": ["points", "q"],
                    "intensity": intensities,
                    "points": numpy.arange(6),
                    "q": q,
                    "q@units": "nm^-1",
                },
            },
            h5f,
        )

    return integrated_patterns_file



from collections.abc import Sequence
import os
import logging
from importlib.metadata import version

from .hdf5_widget import Hdf5Widget
from .utils import (
    generateInputs,
    FilenameCompleterLineEdit,
    get_scan_url,
    ScanEntry,
)

import PyQt5.QtCore  # noqa: F401; Needed to force PyQt5
import qtawesome
from silx.gui import qt, icons

from .constants import FILE_EXTENSIONS

from ..utils import (
    FLATFIELD_DEFAULT_DIR,
    NEWFLAT_FILENAME,
    OLDFLAT_FILENAME,
)

from .jobs_list_widget import JobsListWidget
from .jobs_progressbar import JobsProgressBar
from .processing_groupbox import ProcessingGroupBox
from .output_groupbox import OutputWidget
from .execution_groupbox import ExecutionGroupBox
from .utils import ExportMode
from .workflow_executor import WorkflowExecutor

_logger = logging.getLogger(__name__)


class MainWindow(qt.QMainWindow):
    """
    This GUI is designed for reprocessing HDF5 scans data with fast azimuthal
    integration using Ewoks workflows (ewoksXRPD)
    """

    _SETTINGS_VERSION_STR: str = "2"
    _WORKFLOW: str = "integrate_with_saving_with_flat.json"
    _WORKFLOW_LOAD_OPTIONS: dict[str, str] = {"root_module": "ewoksid31.workflows"}

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__(parent)
        self.setWindowTitle("pyFAI with EWOKS - ID31")
        self.resize(1000, 800)

        self._executor = WorkflowExecutor()
        self._executor.finished.connect(self._executorFinished)

        self.statusBar().addPermanentWidget(JobsProgressBar(self._executor))

        self._executionDockWidget = qt.QDockWidget("Processing Queue")
        self._executionDockWidget.setFeatures(
            qt.QDockWidget.DockWidgetFeature.NoDockWidgetFeatures
        )
        self._executionDockWidget.setWidget(JobsListWidget(self._executor))
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._executionDockWidget)

        # Set paths
        self._outputWidget = OutputWidget()
        self._defaultDirectory = os.getcwd()
        self._flatFieldDirName = FLATFIELD_DEFAULT_DIR

        # Central Layout setup
        centralWidget = qt.QWidget(self)
        self.setCentralWidget(centralWidget)
        mainLayout = qt.QVBoxLayout(centralWidget)

        # HDF5 Viewer
        self._hdf5Widget = Hdf5Widget()
        mainLayout.addWidget(self._hdf5Widget)

        # Menu and Toolbar setup
        self._setupMenuAndToolBar()
        configLayout = qt.QVBoxLayout()

        # pyFAI Config Section
        pyFaiGroupBox = qt.QGroupBox("PyFAI Configuration")
        pyFaiLayout = qt.QGridLayout()

        self._configFileLineEdit = FilenameCompleterLineEdit()
        self._configFileLineEdit.setPlaceholderText("/path/to/pyfai_config.json")

        loadConfigFileButton = qt.QPushButton("Open...")
        loadConfigFileButton.setIcon(icons.getQIcon("document-open"))
        loadConfigFileButton.clicked.connect(self._loadConfigFileButtonClicked)

        pyFaiLayout.addWidget(self._configFileLineEdit, 0, 0, 1, 1)
        pyFaiLayout.addWidget(loadConfigFileButton, 0, 1, 1, 1)

        pyFaiGroupBox.setLayout(pyFaiLayout)
        configLayout.addWidget(pyFaiGroupBox)

        # Processing Options Section
        self._processingGroupBox = ProcessingGroupBox()
        configLayout.addWidget(self._processingGroupBox)

        # Output Section
        configLayout.addWidget(self._outputWidget)

        # Execution Section
        self._executionGroupBox = ExecutionGroupBox()

        configLayout.addWidget(self._executionGroupBox)
        configWidget = qt.QWidget()
        configWidget.setLayout(configLayout)
        configWidget.setSizePolicy(qt.QSizePolicy.Preferred, qt.QSizePolicy.Fixed)

        scrollArea = qt.QScrollArea()
        scrollArea.setWidget(configWidget)
        scrollArea.setWidgetResizable(True)
        scrollArea.setHorizontalScrollBarPolicy(
            qt.Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        scrollArea.setMaximumHeight(configWidget.sizeHint().height() + 2)

        mainLayout.addWidget(scrollArea)

        runButton = qt.QPushButton("Run")
        runButton.setIcon(icons.getQIcon("next"))
        runButton.clicked.connect(self._runButtonClicked)
        runButton.setToolTip("Run the processing workflow on the selected host.")
        runButton.setMinimumSize(runButton.sizeHint() * 2)
        runButton.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
        mainLayout.addWidget(runButton, alignment=qt.Qt.AlignCenter)

    def getDefaultDirectory(self) -> str:
        """
        Return the default directory defined by the user.
        """
        return self._defaultDirectory

    def setDefaultDirectory(self, path: str) -> None:
        """
        Sets the default directory used when opening the QFileDialog.
        """
        self._defaultDirectory = path

    def _showHelpDialogClicked(self):
        """
        Display the Help dialog when the Help button is clicked.
        """
        helpDialog = qt.QMessageBox(self)
        helpDialog.setWindowTitle("Help")
        helpDialog.setIcon(qt.QMessageBox.Information)

        helpDialog.setTextFormat(qt.Qt.RichText)
        helpDialog.setText(
            f"<b>Welcome to the pyFAI with EWOKS application help (version {version('ewoksid31')}).</b><br>"
            "<br>How to get started?<br>"
            "<ul>"
            f"<li>Load raw data files {FILE_EXTENSIONS}.</li>"
            "<li>Load the pyFAI configuration file.</li>"
            "<li>Select an output directory to save the processed data.</li>"
            "<li>Click Run to execute the workflow.</li>"
            "</ul>"
            "<br>For more information, visit:<br>"
            '<a href="https://ewoksid31.readthedocs.io/en/latest/">Ewoksid31 Documentation</a><br>'
            '<a href="https://confluence.esrf.fr/display/ID31KB/GUI+for+reprocessing+XRPD+data">ID31 GUI Confluence Page</a>'
        )

        helpDialog.setStandardButtons(qt.QMessageBox.Ok)
        helpDialog.exec_()

    def _setupMenuAndToolBar(
        self,
    ) -> None:
        """
        Setup of the main menu and toolbar with actions for file handling.
        """
        menuBar = self.menuBar()
        fileMenu = menuBar.addMenu("File")

        openAction = qt.QAction(icons.getQIcon("document-open"), "Open...", self)
        openAction.triggered.connect(self._openActionTriggered)
        openAction.setShortcut(qt.QKeySequence.StandardKey.Open)
        reloadAction = qt.QAction(icons.getQIcon("view-refresh"), "Reload", self)
        reloadAction.triggered.connect(self._hdf5Widget.reloadSelected)
        reloadAction.setShortcut(qt.QKeySequence.StandardKey.Refresh)
        clearAction = qt.QAction(icons.getQIcon("remove"), "Clear", self)
        clearAction.triggered.connect(self._hdf5Widget.clearFiles)
        clearAction.setShortcut(qt.QKeySequence.StandardKey.Delete)

        fileMenu.addActions([openAction, reloadAction, clearAction])

        toolBar = self.addToolBar("File Actions")
        toolBar.addActions([openAction, reloadAction, clearAction])
        toolBar.setMovable(False)
        self.setContextMenuPolicy(qt.Qt.PreventContextMenu)
        toolBar.setFloatable(False)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Preferred)
        toolBar.addWidget(spacer)

        helpAction = qt.QAction(
            self.style().standardIcon(qt.QStyle.SP_TitleBarContextHelpButton), "", self
        )
        font = helpAction.font()
        font.setBold(True)
        helpAction.setFont(font)
        helpAction.setToolTip("About this app")
        helpAction.triggered.connect(self._showHelpDialogClicked)
        toolBar.addAction(helpAction)

        jobsListVisibleAction = qt.QAction(
            qtawesome.icon("fa6s.list-check", scale_factor=0.75), "", self
        )
        jobsListVisibleAction.setToolTip("Show/Hide the processing queue panel")
        jobsListVisibleAction.setCheckable(True)
        jobsListVisibleAction.setChecked(self._executionDockWidget.isVisibleTo(self))
        jobsListVisibleAction.toggled.connect(self._executionDockWidget.setVisible)
        self._executionDockWidget.visibilityChanged.connect(
            jobsListVisibleAction.setChecked
        )
        toolBar.addAction(jobsListVisibleAction)

    def addRawDataFile(self, fileName: str) -> None:
        """
        Proxy method to add a new file to the HDF5 tree viewer.
        """
        if not fileName or not os.path.isfile(fileName):
            qt.QMessageBox.warning(
                self, "Invalid File Format", "The selected file does not exist."
            )
            return
        if not fileName.endswith(FILE_EXTENSIONS):
            qt.QMessageBox.warning(
                self,
                "Invalid File Format",
                f"Please select a valid HDF5 or NEXUS file {FILE_EXTENSIONS}.",
            )
            return
        self._hdf5Widget.addFile(fileName)
        self.setDefaultDirectory(os.path.dirname(fileName))

    def _openActionTriggered(self) -> None:
        """
        Add Raw data as HDF5 file without cleaning the tree viewer.
        """
        nameFilter = " ".join(f"*{ext}" for ext in FILE_EXTENSIONS)
        fileName, _ = qt.QFileDialog.getOpenFileName(
            self,
            "Add RAW data file",
            self.getDefaultDirectory(),
            f"HDF5 files ({nameFilter});;All files (*)",
        )
        if fileName:
            self.addRawDataFile(fileName)

    def _loadConfigFileButtonClicked(self) -> None:
        """
        Choose and import JSON or PONI config file.
        """
        currentDir = (
            os.path.dirname(self.getConfigFilePath()) or self.getDefaultDirectory()
        )
        filePath, _ = qt.QFileDialog.getOpenFileName(
            self,
            "Open config file",
            currentDir,
            "JSON or PONI files (*.json *.poni);;All files (*)",
        )
        if filePath:
            self.setConfigFilePath(filePath)
        else:
            self.setConfigFilePath("")
            _logger.info("No config file chosen.")

    def isWorkflowRunning(self) -> bool:
        return self._executor.isJobRunning()

    def getFlatFieldDirName(self) -> str:
        """Returns the flat-field directory where to find flats.mat and oldflats.mat"""
        return self._flatFieldDirName

    def setFlatFieldDirName(self, path: str):
        """Set the directory where to find flats.mat and oldflats.mat"""
        self._flatFieldDirName = os.path.abspath(path)

    def getConfigFilePath(self) -> str:
        """
        Returns the current configuration file path from the line edit.
        """
        return self._configFileLineEdit.text().strip()

    def setConfigFilePath(self, path: str) -> None:
        """
        Update the configuration file path in the line edit.
        """
        self._configFileLineEdit.setText(path)

    def _getParameters(self, scan: ScanEntry, unit: str) -> dict | None:
        """
        Generates parameters to execute workflow for a given scan.

        Args:
            datasetFilename: Filename of the HDF5 file containing the scan
            scanNumber: Number of the scan.
            unit: Selected unit.

        Returns:
            Dictionnary of parameters for the workflow, or None if failed.
        """

        return generateInputs(
            scan=scan,
            newFlat=os.path.join(self.getFlatFieldDirName(), NEWFLAT_FILENAME),
            oldFlat=os.path.join(self.getFlatFieldDirName(), OLDFLAT_FILENAME),
            pyfaiConfigFile=self.getConfigFilePath(),
            pyfaiMethod=self._processingGroupBox.getIntegrationMethod(),
            monitorName=self._processingGroupBox.getMonitorName(),
            referenceCounts=1,
            detectorName=self._processingGroupBox.getDetectorName(),
            outputDirectory=self._outputWidget.getOutputDirName(),
            sigmaClippingThreshold=self._processingGroupBox.getSigmaClippingThreshold(),
            exportMode=self._outputWidget.getExportMode(),
            unit=unit,
            rotName=self._outputWidget.getTomoRotName(),
            yName=self._outputWidget.getTomoYName(),
        )

    def _prepareScans(self, scans: Sequence[ScanEntry]) -> list[dict]:
        """
        Prepares parameters for each unique selected scan and selected unit.
        """
        inputParameters = list()
        selectedUnits = self._processingGroupBox.getSelectedUnits()
        for scan in scans:
            for unit in selectedUnits:
                workflowParameters = self._getParameters(scan, unit)
                if not workflowParameters:
                    _logger.warning(
                        f"Skipping scan {scan.filename}::{scan.number} due to missing parameters."
                    )
                    continue

                inputParameters.append(workflowParameters)

        return inputParameters

    def _processScans(self, scans: Sequence[ScanEntry], local: bool = True) -> None:
        """
        Execute workflow and save HDF5 data and JSON workflow ewoks file from a single selected scan
        and selected unit.

        Executing in a separate thread.

        Args:
            scans: Selected scans filename and number
            local: Whether to execute locally or submit to Ewoks worker.
        """
        if not scans:
            return

        self.statusBar().showMessage("Processing selected scans...")

        for inputParameters in self._prepareScans(scans):
            scanFilename, scanNumber = get_scan_url(inputParameters)
            _logger.info(f"Submit workflow for {scanFilename}::{scanNumber}")
            self._executor.submit(
                local,
                graph=self._WORKFLOW,
                load_options=self._WORKFLOW_LOAD_OPTIONS,
                **inputParameters,
            )

    def _executorFinished(self) -> None:
        self.statusBar().showMessage("Processing finished")

    def _runButtonClicked(self) -> None:
        """
        Handle the run button click. Validate inputs, adjust output directory and process scans based on the selected execution mode.
        """
        selectedScans = self._hdf5Widget.getFilteredSelectedScans(
            self._processingGroupBox.getDetectorName()
        )
        userSelectedScans = self._hdf5Widget.getUserSelectedScans()
        invalidScans = userSelectedScans - selectedScans

        if invalidScans:
            for invalidScan in sorted(invalidScans):
                _logger.warning(
                    f"Ignoring scan {os.path.basename(invalidScan.filename)}::/{invalidScan.number}.1"
                )

        if not self._validateInputParameters(selectedScans):
            return

        if not self._adjustOutputDirectory():
            self.statusBar().showMessage("Process canceled by user.")
            return

        self._processScans(
            tuple(sorted(selectedScans)),
            local=self._executionGroupBox.isLocalExecution(),
        )

    def _getValidationErrors(
        self,
        rawDataLoaded: bool,
        selectedScans: set[ScanEntry],
        configFileLoaded: bool,
        unitSelected: bool,
    ) -> list[str]:
        """
        Generate a list of validation error messages based on the current state.

        Args:
            rawDataLoaded: Whether raw data is loaded.
            selectedScans: List of selected scans.
            configFileLoaded: Whether a configuration file is loaded.
            unitSelected: Whether at least one radial unit is selected.

        Returns:
            A list of error messages if validation fails, otherwise an empty list.
        """
        errors = list()

        if not rawDataLoaded:
            errors.append("No raw data file loaded.")
        if not selectedScans:
            errors.append("No scan selected for processing.")
        if not configFileLoaded:
            errors.append("No pyFAI config file loaded.")
        if not unitSelected:
            errors.append("No radial unit selected.")
        return errors

    def _showWarningMessage(self, errorMessages: list[str]) -> None:
        """
        Show a warning message box if prerequisites for processing are not met.
        """
        warningMessageBox = qt.QMessageBox(self)
        warningMessageBox.setWindowTitle("Workflow cannot be excecuted")
        warningMessageBox.setIcon(qt.QMessageBox.Warning)
        warningMessageBox.setText("\n".join(errorMessages))
        warningMessageBox.exec_()

    def _validateInputParameters(self, selectedScans: set) -> bool:
        """
        Validates that all necessary inputs are present before starting processing.

        Returns:
            True if inputs are valid, False otherwise.
        """
        rawDataLoaded = not self._hdf5Widget.isEmpty()
        configFileLoaded = bool(self.getConfigFilePath())
        unitSelected = len(self._processingGroupBox.getSelectedUnits()) != 0

        errorMessages = self._getValidationErrors(
            rawDataLoaded, selectedScans, configFileLoaded, unitSelected
        )

        if errorMessages:
            self.statusBar().showMessage(" ".join(errorMessages))
            _logger.warning(f"Processing cannot start: {' '.join(errorMessages)}")
            self._showWarningMessage(errorMessages)
            return False
        return True

    def _showConfirmationMessage(self, outputDirectory: str) -> bool:
        """
        Show a confirmation message box when creating a new directory for output file.

        Returns True is the user confirms, otherwise False.
        """
        confirmationMessageBox = qt.QMessageBox(self)
        confirmationMessageBox.setWindowTitle("Overwrite Output Files")
        confirmationMessageBox.setIcon(qt.QMessageBox.Question)
        confirmationMessageBox.setText("The selected output directory already exists.")
        confirmationMessageBox.setInformativeText(
            f"Do you want to overwrite existing files in: {outputDirectory} ?\nUnchanged files will be preserved."
        )
        confirmationMessageBox.setStandardButtons(
            qt.QMessageBox.Ok | qt.QMessageBox.Cancel
        )
        userResponse = confirmationMessageBox.exec_()
        return userResponse == qt.QMessageBox.Ok

    def _adjustOutputDirectory(self) -> bool:
        """
        Use the selected output directory directly, without creating a new one.
        Ask for confirmation if the folder already exists.

        Returns:
            True if the process should continue, False if canceled.
        """
        currentOutputDir = self._outputWidget.getOutputDirName()
        if os.path.exists(currentOutputDir):
            if not self._showConfirmationMessage(currentOutputDir):
                _logger.info("Process canceled by the user.")
                return False
            _logger.info(
                f"Processed data will be written in existing directory: {currentOutputDir}"
            )

        return True

    def loadSettings(self) -> None:
        """
        Load user settings.
        """
        settings = qt.QSettings()

        if settings.value("version") != self._SETTINGS_VERSION_STR:
            _logger.info("Setting version mismatch. Clearing settings.")
            settings.clear()
            return

        geometry = settings.value("mainWindow/geometry")
        if geometry:
            self.restoreGeometry(geometry)

        defaultDirectory = settings.value("input/inputDirectory", None)
        if defaultDirectory and os.path.isdir(defaultDirectory):
            self.setDefaultDirectory(defaultDirectory)
        self.setConfigFilePath(settings.value("config/configFile", ""))
        self._outputWidget.setOutputDirName(
            settings.value("output/outputDirectory", "")
        )

        self._processingGroupBox.loadSettings()

        self._outputWidget.setTomoRotName(settings.value("output/tomo/rot_name", "nth"))
        self._outputWidget.setTomoYName(settings.value("output/tomo/y_name", "ny"))

        self._outputWidget.setExportMode(
            ExportMode(
                settings.value("output/exportMode", ExportMode.DISABLED.value, type=int)
            )
        )

        self._executionGroupBox.setLocalExecution(
            settings.value("run/localExecution", True, type=bool)
        )

        self._hdf5Widget.loadSettings()

    def closeEvent(self, event: qt.QCloseEvent) -> None:
        """
        Save user settings on application close.
        """
        if self.isWorkflowRunning():
            button = qt.QMessageBox.warning(
                self,
                "Processing is running",
                "<b>Do you want to quit anyway?</b><br><br>This might corrupt local output files.",
                qt.QMessageBox.Yes | qt.QMessageBox.No,
                qt.QMessageBox.No,
            )
            if button != qt.QMessageBox.Yes:
                event.ignore()
                return
            self._executor.shutdown(wait=False, cancelFutures=True)

        settings = qt.QSettings()

        settings.setValue("version", self._SETTINGS_VERSION_STR)
        settings.setValue("mainWindow/geometry", self.saveGeometry())

        settings.setValue("input/inputDirectory", self.getDefaultDirectory())

        settings.setValue("config/configFile", self.getConfigFilePath())

        self._processingGroupBox.saveSettings()

        settings.setValue("output/tomo/rot_name", self._outputWidget.getTomoRotName())
        settings.setValue("output/tomo/y_name", self._outputWidget.getTomoYName())

        settings.setValue(
            "output/outputDirectory", self._outputWidget.getOutputDirName()
        )
        settings.setValue("output/exportMode", self._outputWidget.getExportMode().value)

        settings.setValue(
            "run/localExecution", self._executionGroupBox.isLocalExecution()
        )

        self._hdf5Widget.saveSettings()



from collections.abc import Generator
import fnmatch
import logging

import hdf5plugin  # noqa
import h5py

from silx.gui import qt
from silx.gui.data.DataViewerFrame import DataViewerFrame
import silx.io

from .hdf5_treeview import Hdf5TreeView
from .selection_filter_groupbox import SelectionFilterGroupBox
from .utils import extractScanNumber, ScanEntry


_logger = logging.getLogger(__name__)


class Hdf5Widget(qt.QWidget):

    def __init__(self) -> None:
        super().__init__()
        self.__loadedFiles: set[str] = set()

        self.__treeview = Hdf5TreeView(self)

        self.__dataViewer = DataViewerFrame(self)
        self.__treeview.h5NodeActivated.connect(self.__dataViewer.setData)

        self.__treeview.setMinimumSize(150, 200)
        self.__dataViewer.setMinimumSize(200, 200)

        splitter = qt.QSplitter(self)
        splitter.addWidget(self.__treeview)
        splitter.addWidget(self.__dataViewer)

        splitter.setCollapsible(0, False)
        splitter.setCollapsible(1, False)

        splitter.setHandleWidth(10)
        splitter.setStretchFactor(1, 1)

        self._selectionFilterGroupBox = SelectionFilterGroupBox(self)

        layout = qt.QVBoxLayout(self)
        layout.setContentsMargins(qt.QMargins())
        layout.addWidget(splitter, stretch=1)
        layout.addWidget(self._selectionFilterGroupBox, stretch=0)

    def addFile(self, filename: str) -> None:
        """
        Add a file to the HDF5 tree viewer.

        Prevents adding duplicate files by checking if the file is already loaded.

        Args:
            filename: Path to the HDF5 file to add.
        """
        if filename in self.__loadedFiles:
            qt.QMessageBox.information(
                self,
                "Duplicate File",
                f"This file {filename} is already loaded.",
            )
            return

        self.__treeview.findHdf5TreeModel().appendFile(filename)
        self.__loadedFiles.add(filename)

    def clearFiles(self) -> None:
        """
        Remove and close all files currently loaded in the HDF5 tree viewer.

        Resets the internal file tracking list.
        """
        self.__treeview.findHdf5TreeModel().clear()
        self.__loadedFiles.clear()

    def getLoadedFiles(self) -> set[str]:
        """
        Get the set of currently loaded files in the HDF5 tree viewer.

        Returns:
            A set containing the file paths of all loaded files.
        """
        return self.__loadedFiles

    def isEmpty(self) -> bool:
        """
        Check if the HDF5 tree viewer contains any loaded files.

        Returns:
            True if no files are currently loaded, False otherwise.
        """
        return len(self.__loadedFiles) == 0

    def reloadSelected(self) -> None:
        """
        Reload the currently selected HDF5 nodes.
        """
        model = self.__treeview.findHdf5TreeModel()
        for obj in tuple(self.__treeview.selectedH5Nodes()):
            if obj.ntype is h5py.File:
                model.synchronizeH5pyObject(obj.h5py_object)

    def _getSelectedScansAndPaths(self) -> Generator[tuple[ScanEntry, str]]:
        for scan in self.__treeview.selectedH5Nodes():
            scanNumber = extractScanNumber(scan.physical_name)
            if scanNumber == -1:
                _logger.warning(
                    f"Skipping scan: Invalid scan number in {scan.physical_name}"
                )
                continue
            scanEntry = ScanEntry(filename=scan.physical_filename, number=scanNumber)
            yield scanEntry, scan.local_name

    def getUserSelectedScans(self) -> set[ScanEntry]:
        """
        Return all scans selected by the user in the silx tree viewer.
        """
        return set(scanEntry for scanEntry, _ in self._getSelectedScansAndPaths())

    def getFilteredSelectedScans(self, detectorName: str) -> set[ScanEntry]:
        """
        Returns user selected scans filtered based on:

        - Scan group name patterns, matched against the user selected HDF5 path
        - Scan command patterns, matched against the HDF5 'title' dataset
        - Presence of the selected detector dataset in the scan
        """
        scanGroupNamePatterns = self._selectionFilterGroupBox.getScanGroupNamePatterns()
        scanCommandPatterns = self._selectionFilterGroupBox.getScanCommandPatterns()

        filteredScans = set()
        for scan, selectedPath in self._getSelectedScansAndPaths():
            if scanGroupNamePatterns:
                scanGroupName = selectedPath.lstrip("/").split("/")[0]
                matchedPatterns = [
                    pattern
                    for pattern in scanGroupNamePatterns
                    if fnmatch.fnmatch(scanGroupName.lower(), f"*{pattern.lower()}*")
                ]
                if not matchedPatterns:
                    continue

            with silx.io.open(scan.filename) as h5f:
                detectorDatasetName = f"{scan.number}.1/measurement/{detectorName}"
                if detectorDatasetName not in h5f:
                    continue

                if scanCommandPatterns:
                    titleDataset = h5f.get(f"{scan.number}.1/title")
                    if titleDataset is None:
                        continue
                    scanTitle = titleDataset.asstr()[()]
                    matchedPatterns = [
                        pattern
                        for pattern in scanCommandPatterns
                        if fnmatch.fnmatch(scanTitle.lower(), f"{pattern.lower()}*")
                    ]
                    if not matchedPatterns:
                        continue

            filteredScans.add(scan)
        return filteredScans

    def loadSettings(self) -> None:
        self._selectionFilterGroupBox.loadSettings()

    def saveSettings(self) -> None:
        self._selectionFilterGroupBox.saveSettings()



from typing import Any, Literal

from concurrent.futures import Future, ThreadPoolExecutor
import copy
from datetime import datetime

import ewoks
from ewoksjob.client.futures import FutureInterface
from silx.gui import qt
import silx.gui.qt.inspect


class JobItem:
    def __init__(
        self, future: Future | FutureInterface, local: bool, kwargs: dict[str, Any]
    ) -> None:
        self._timestamp = datetime.now()
        self._future = future
        self._local = local
        self._kwargs: dict[str, Any] = copy.deepcopy(kwargs)
        self._state: Literal["pending", "running", "done"] = "pending"

    def getFuture(self) -> Future | FutureInterface:
        return self._future

    def isLocal(self) -> bool:
        return self._local

    def getArguments(self) -> dict[str, Any]:
        return copy.deepcopy(self._kwargs)

    def getTimestamp(self) -> datetime:
        return self._timestamp

    def _isStateUpdated(self) -> bool:
        if self._state == "done":
            return False

        if self._state == "pending" and self._future.running():
            self._state = "running"
            return True

        if self._future.done():
            self._state = "done"
            return True
        return False


class WorkflowExecutor(qt.QObject):
    """Execute workflows locally or remotely and reports submitted job activity"""

    jobSubmitted = qt.Signal(JobItem)

    jobChanged = qt.Signal(JobItem)
    """Signal emitted when a job has changed from pending to running or running to done or cancelled"""

    finished = qt.Signal()
    """Signal emitted when all submitted jobs are finished (failed, cancelled or succeeded)"""

    def __init__(self, parent: qt.QObject | None = None) -> None:
        super().__init__(parent)
        self._timer = qt.QTimer(self)
        self._timer.setInterval(1000)  # milliseconds
        self._timer.timeout.connect(self._updateItems)
        self._items: set[JobItem] = set()
        self._localExecutor = ThreadPoolExecutor(
            max_workers=1, thread_name_prefix=self.__class__.__name__
        )

    def _remoteShutdown(self, wait: bool, cancelFutures: bool) -> None:
        remoteFutures = set(
            item.getFuture() for item in self._items if not item.isLocal()
        )
        if cancelFutures:
            for future in remoteFutures:
                future.cancel()
        if wait:
            for future in remoteFutures:
                try:
                    future.result()
                except Exception:  # nosec: B110
                    pass

    def shutdown(self, wait: bool = True, *, cancelFutures: bool = False) -> None:
        self._localExecutor.shutdown(wait=wait, cancel_futures=cancelFutures)
        self._remoteShutdown(wait=wait, cancelFutures=cancelFutures)
        self._updateItems()

    def __del__(self):
        self.shutdown(wait=False, cancelFutures=True)

    def submit(self, local: bool, /, **kwargs) -> JobItem:
        # Use a copy: execute_graph modify its graph argument
        kwargs_copy = copy.deepcopy(kwargs)
        if local:
            future = self._localExecutor.submit(ewoks.execute_graph, **kwargs_copy)
        else:
            future = ewoks.submit_graph(**kwargs_copy)

        item = JobItem(future, local, kwargs)
        self._items.add(item)
        if not self._timer.isActive():
            self._timer.start()
        self.jobSubmitted.emit(item)
        return item

    def _updateItems(self):
        for item in self._items.copy():
            if item._isStateUpdated():
                self.jobChanged.emit(item)
            if item.getFuture().done():
                self._items.discard(item)

        if not self._items:
            # Since this is called from __del__, make sure QObjects are valid
            if silx.gui.qt.inspect.isValid(self):
                if silx.gui.qt.inspect.isValid(self._timer):
                    self._timer.stop()
                self.finished.emit()

    def isJobRunning(self) -> bool:
        return bool(self._items)



from silx.gui import qt


class SelectionFilterGroupBox(qt.QGroupBox):
    """GroupBox for configuring the filtering of selected scans"""

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__("Filter Selection", parent)
        self.setCheckable(True)
        self.setChecked(False)

        layout = qt.QHBoxLayout(self)

        layout.addWidget(qt.QLabel("Scan group name patterns:"))
        self._scanGroupNamePatternsLineEdit = qt.QLineEdit()
        self._scanGroupNamePatternsLineEdit.setToolTip(
            "Comma-separated list of patterns to filter scan group names in selection (e.g. WAXS, SAXS)"
        )
        self._scanGroupNamePatternsLineEdit.setPlaceholderText("scangroupname")
        layout.addWidget(self._scanGroupNamePatternsLineEdit)

        layout.addWidget(qt.QLabel("Scan command patterns:"))
        self._scanCommandPatternsLineEdit = qt.QLineEdit()
        self._scanCommandPatternsLineEdit.setToolTip(
            "Comma-separated list of patterns to filter scan title in selection (e.g. ct, ascan)"
        )
        self._scanCommandPatternsLineEdit.setPlaceholderText("scancommand")
        layout.addWidget(self._scanCommandPatternsLineEdit)

    def getScanGroupNamePatterns(self) -> tuple[str, ...]:
        """Return patterns to match scan group names (e.g. 'WAXS', 'SAXS')."""
        if not self.isChecked():
            return ()
        text = self._scanGroupNamePatternsLineEdit.text()
        patterns = [pattern.strip() for pattern in text.split(",")]
        return tuple(p for p in patterns if p)

    def getScanCommandPatterns(self) -> tuple[str, ...]:
        """Return patterns to match scan commands (e.g. 'ct', 'ascan')."""
        if not self.isChecked():
            return ()
        text = self._scanCommandPatternsLineEdit.text()
        patterns = [pattern.strip() for pattern in text.split(",")]
        return tuple(p for p in patterns if p)

    def loadSettings(self) -> None:
        settings = qt.QSettings()
        self.setChecked(settings.value("filterSelection/checked", False, type=bool))
        self._scanGroupNamePatternsLineEdit.setText(
            settings.value("filterSelection/scanGroupNamePatterns", "", type=str)
        )
        self._scanCommandPatternsLineEdit.setText(
            settings.value("filterSelection/scanCommandPatterns", "", type=str)
        )

    def saveSettings(self) -> None:
        settings = qt.QSettings()
        settings.setValue("filterSelection/checked", self.isChecked())
        settings.setValue(
            "filterSelection/scanGroupNamePatterns",
            self._scanGroupNamePatternsLineEdit.text(),
        )
        settings.setValue(
            "filterSelection/scanCommandPatterns",
            self._scanCommandPatternsLineEdit.text(),
        )



from collections import defaultdict
from typing import Any


FLATFIELD_DEFAULT_DIR = "/data/id31/inhouse/P3/"
NEWFLAT_FILENAME = "flats.mat"
OLDFLAT_FILENAME = "flats_old.mat"


def print_inputs(inputs: list[dict[str, Any]]) -> None:
    """Pretty print a list of ewoks task inputs"""
    tasks_args = defaultdict(list)
    for input in inputs:
        task_key = input.get("task_identifier", "")
        task_id = input.get("id")
        if task_id:
            task_key += f" <id={task_id}"
        task_label = input.get("label")
        if task_label:
            task_key += f' "{task_label}"'
        tasks_args[task_key].append((input["name"], input["value"]))

    for task_key in sorted(tasks_args.keys()):
        print(f"{task_key}:")
        for name, value in tasks_args[task_key]:
            print(f"- {name}: {value}")
        print("")



from ewoksutils.task_utils import task_inputs

import pytest

from ..app.integrate.workflow_executor import WorkflowExecutor


SUM_TASK_WORKFLOW = {
    "graph": {"id": "test"},
    "nodes": [
        {
            "id": "task",
            "task_type": "class",
            "task_identifier": "ewokscore.tests.examples.tasks.sumtask.SumTask",
        },
    ],
}


def test_success(qtbot):
    """Submit one workflow that succeed"""
    executor = WorkflowExecutor()

    inputs = task_inputs(inputs={"a": 1, "b": 2, "delay": 2})

    with qtbot.waitSignal(executor.jobSubmitted, timeout=1000) as blocker:
        job = executor.submit(True, graph=SUM_TASK_WORKFLOW, inputs=inputs)
    assert job is blocker.args[0]

    with qtbot.waitSignal(executor.finished, timeout=20000):
        executor.shutdown(wait=True, cancelFutures=False)

    assert not executor.isJobRunning()

    assert job.isLocal()

    arguments = job.getArguments()
    assert arguments["graph"] == SUM_TASK_WORKFLOW
    assert arguments["inputs"] == inputs

    future = job.getFuture()
    assert future.done()
    assert future.result() == {"result": 3}


def test_cancel(qtbot):
    """Submit 2 workflows and cancel second one"""
    executor = WorkflowExecutor()

    inputs = task_inputs(inputs={"a": 1, "b": 2, "delay": 2})

    job1 = executor.submit(True, graph=SUM_TASK_WORKFLOW, inputs=inputs)
    job2 = executor.submit(True, graph=SUM_TASK_WORKFLOW, inputs=inputs)

    assert executor.isJobRunning()

    future2 = job2.getFuture()
    assert future2.cancel()

    with qtbot.waitSignal(executor.finished, timeout=20000):
        executor.shutdown(wait=True, cancelFutures=False)

    assert job1.getFuture().done()
    assert future2.cancelled()


def test_failed(qtbot):
    """Submit one workflow that raises an exception"""
    executor = WorkflowExecutor()

    workflow = {
        "graph": {"id": "test"},
        "nodes": [
            {
                "id": "task",
                "task_type": "class",
                "task_identifier": "ewokscore.tests.examples.tasks.errorsumtask.ErrorSumTask",
            },
        ],
    }
    inputs = task_inputs(inputs={"raise_error": True})

    job = executor.submit(True, graph=workflow, inputs=inputs)

    with qtbot.waitSignal(executor.finished, timeout=20000):
        executor.shutdown(wait=True, cancelFutures=False)

    future = job.getFuture()
    assert future.done()
    with pytest.raises(RuntimeError):
        future.result()
    assert isinstance(future.exception(), RuntimeError)



from dataclasses import dataclass
import logging

import os
import h5py
import numpy
from pydantic import Field
from silx.io import h5py_utils
from silx.io.nxdata import NXdata
from silx.io.url import DataUrl

from ewokscore import Task
from ewokscore.model import BaseInputModel
from ewoksxrpd.tasks.utils.ascii_utils import ensure_parent_folder


logger = logging.getLogger(__name__)


class Inputs(BaseInputModel):
    scan_entry_url: str = Field(
        description="HDF5 URL to the NXentry group containing the 'measurement' group",
        examples=["/path/to/raw_data/file.f5::/2.1"],
    )
    rot_name: str = Field(
        description="Dataset name in 'measurement' for rotation angles",
        examples=["nth"],
    )
    y_name: str = Field(
        description="Dataset name in 'measurement' for horizontal positions",
        examples=["ny"],
    )
    nxdata_url: str = Field(
        description="HDF5 URL to the NXdata group containing integrated patterns",
        examples=["/path/to/processed_data/file.f5::/2.1/p3_integrate/integrated"],
    )
    output_filename: str = Field(
        description="Path to the output HDF5 file",
        examples=["/path/to/processed_data/tomo.f5"],
    )
    enabled: bool = True
    overwrite: bool = False
    retry_timeout: float | None = Field(
        default=None,
        description="Duration in seconds for which to try to open HDF5 files",
    )
    retry_period: float | None = Field(
        default=None,
        description="Amount of time in seconds between each attempt to open HDF5 files",
    )


@dataclass
class IntegratedPatterns:
    """Store multiple pyFAI integrated patterns"""

    radial: numpy.ndarray
    radial_name: str
    radial_units: str
    intensities: numpy.ndarray


class SaveNexusPatternsAsId31TomoHdf5(Task, input_model=Inputs, output_names=["filename"]):  # type: ignore[call-arg]
    """Save integrated XRD patterns into an HDF5 file compatible with ID31 tomography workflow.

    Output:

    - filename (str): Path to the written HDF5 file, or empty string if skipped.
    """

    def run(self):
        inputs = Inputs(**self.get_input_values())

        if not inputs.enabled:
            logger.info(
                f"Task {self.__class__.__qualname__} is disabled: No file is saved"
            )
            self.outputs.filename = ""
            return

        ensure_parent_folder(inputs.output_filename)

        scan_entry_url = DataUrl(inputs.scan_entry_url)
        with h5py_utils.open_item(
            scan_entry_url.file_path(),
            scan_entry_url.data_path(),
            retry_timeout=inputs.retry_timeout,
            retry_period=inputs.retry_period,
        ) as scan_entry:
            measurement = scan_entry["measurement"]

            th_angles_dataset = measurement[inputs.rot_name]
            th_angles = th_angles_dataset[()]
            th_angles_units = th_angles_dataset.attrs.get("units", None)

            y_positions_dataset = measurement[inputs.y_name]
            y_positions = y_positions_dataset[()]
            y_positions_units = y_positions_dataset.attrs.get("units", None)

        nxdata_url = DataUrl(inputs.nxdata_url)
        with h5py_utils.open_item(
            nxdata_url.file_path(),
            nxdata_url.data_path(),
            retry_timeout=inputs.retry_timeout,
            retry_period=inputs.retry_period,
        ) as group:
            patterns = _read_nexus_integrated_patterns(group)

        if inputs.overwrite:
            mode = "w"
        else:
            mode = "a"

        if os.path.exists(inputs.output_filename) and not inputs.overwrite:
            raise FileExistsError(
                f"{inputs.output_filename} already exists. Use overwrite=True to replace it."
            )

        with h5py.File(inputs.output_filename, mode) as h5f:
            h5f["XRD"] = patterns.intensities

            h5f[patterns.radial_name] = patterns.radial
            if patterns.radial_units:
                h5f[patterns.radial_name].attrs["units"] = patterns.radial_units

            h5f["y"] = y_positions
            if y_positions_units is not None:
                h5f["y"].attrs["units"] = y_positions_units

            h5f["th"] = th_angles
            if th_angles_units is not None:
                h5f["th"].attrs["units"] = th_angles_units

        self.outputs.filename = inputs.output_filename


def _read_nexus_integrated_patterns(group: h5py.Group) -> IntegratedPatterns:
    """Read integrated patterns from a HDF5 NXdata group.

    It reads from both single (1D signal) or multi (2D signal) NXdata.
    """
    nxdata = NXdata(group)
    if not nxdata.is_valid:
        raise RuntimeError(
            f"Cannot parse NXdata group: {group.file.filename}::{group.name}"
        )
    if not (nxdata.signal_is_1d or nxdata.signal_is_2d):
        raise RuntimeError(
            f"Signal is not a 1D or 2D dataset: {group.file.filename}::{group.name}"
        )

    if nxdata.axes[-1] is None:
        radial = numpy.arange(nxdata.signal.shape[-1])
        radial_name = "radial"
        radial_units = ""
    else:
        axis_dataset = nxdata.axes[-1]
        radial = axis_dataset[()]
        radial_name = nxdata.axes_dataset_names[-1]
        radial_units = axis_dataset.attrs.get("units", "")

    intensities = numpy.atleast_2d(nxdata.signal)

    return IntegratedPatterns(radial, radial_name, radial_units, intensities)



from silx.gui import qt
from .constants import MONITOR_NAMES, DETECTOR_NAMES, INTEGRATION_METHODS


class ProcessingGroupBox(qt.QGroupBox):
    def __init__(self):
        super().__init__("Processing")

        layout = qt.QGridLayout(self)

        spacer = qt.QSpacerItem(
            20, 10, qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        layout.addItem(spacer, 0, 4, 2, 4)

        detectorLabel = qt.QLabel("Detector:")
        self._detectorComboBox = qt.QComboBox()
        self._detectorComboBox.addItems(DETECTOR_NAMES)
        self._detectorComboBox.setCurrentText("p3")
        self._detectorComboBox.setToolTip("Select the detector type.")
        self._detectorComboBox.activated.connect(self._onDetectorChanged)

        layout.addWidget(detectorLabel, 0, 0, 1, 1, qt.Qt.AlignLeft)
        layout.addWidget(self._detectorComboBox, 0, 1, 1, 2)

        monitorNameLabel = qt.QLabel("Monitor:")
        self._monitorNameComboBox = qt.QComboBox()
        self._monitorNameComboBox.addItem("-")
        self._monitorNameComboBox.addItems(MONITOR_NAMES)
        self._monitorNameComboBox.setCurrentText("scaled_mondio")
        self._monitorNameComboBox.setToolTip(
            "Select the monitor name for normalization."
        )
        self._monitorNameComboBox.setItemData(
            0, "Disable normalization", qt.Qt.ToolTipRole
        )

        layout.addWidget(monitorNameLabel, 1, 0, 1, 1, qt.Qt.AlignLeft)
        layout.addWidget(self._monitorNameComboBox, 1, 1, 1, 2)

        self._sigmaClipCheckBox = qt.QCheckBox("Sigma clipping threshold:")
        self._sigmaClipCheckBox.setChecked(False)
        self._sigmaClipCheckBox.setToolTip(
            "Check to enable sigma clipping and set threshold."
        )

        self._sigmaClipThresholdSpinBox = qt.QDoubleSpinBox()
        self._sigmaClipThresholdSpinBox.setSingleStep(0.1)
        self._sigmaClipThresholdSpinBox.setRange(0.1, 10.0)
        self._sigmaClipThresholdSpinBox.setValue(3.0)
        self._sigmaClipThresholdSpinBox.setDecimals(1)
        self._sigmaClipThresholdSpinBox.setEnabled(False)

        self._sigmaClipCheckBox.toggled.connect(
            self._sigmaClipThresholdSpinBox.setEnabled
        )
        self._sigmaClipCheckBox.toggled.connect(self._sigmaClipCheckBoxToggled)

        layout.addWidget(self._sigmaClipCheckBox, 2, 0, 1, 1, qt.Qt.AlignLeft)
        layout.addWidget(self._sigmaClipThresholdSpinBox, 2, 1, 1, 2, qt.Qt.AlignLeft)

        integrationMethodLabel = qt.QLabel("Integration method:")
        self._integrationMethodComboBox = qt.QComboBox()
        self._integrationMethodComboBox.addItems(INTEGRATION_METHODS)
        self._integrationMethodComboBox.setCurrentText("no_histogram_cython")
        self._integrationMethodComboBox.setToolTip(
            "Select the pyFAI integration method."
        )

        layout.addWidget(integrationMethodLabel, 3, 0, 1, 1, qt.Qt.AlignLeft)
        layout.addWidget(self._integrationMethodComboBox, 3, 1, 1, 2)

        radialUnitLabel = qt.QLabel("Radial unit:")
        self._qCheckBox = qt.QCheckBox("q (scattering vector)")
        self._qUnitsComboBox = qt.QComboBox()
        self._qUnitsComboBox.addItems(("Å⁻¹", "nm⁻¹"))
        self._2thCheckBox = qt.QCheckBox("2theta (degree)")
        self._qUnitsComboBox.setEnabled(False)
        self._qCheckBox.toggled.connect(self._qUnitsComboBox.setEnabled)

        layout.addWidget(radialUnitLabel, 4, 0, 1, 1, qt.Qt.AlignLeft)
        layout.addWidget(self._qCheckBox, 4, 1, 1, 1, qt.Qt.AlignLeft)
        layout.addWidget(self._qUnitsComboBox, 4, 2, 1, 1, qt.Qt.AlignRight)
        layout.addWidget(self._2thCheckBox, 5, 1, 1, 1, qt.Qt.AlignLeft)

    def getDetectorName(self) -> str:
        return self._detectorComboBox.currentText()

    def setDetectorName(self, value: str):
        self._detectorComboBox.setCurrentText(value)

    def _onDetectorChanged(self) -> None:
        """
        Prompt the user to verify the configuration file after changing the detector.

        Ensures that the selected config file in the 'pyFAI config' section is compatible with the newly selected detector.
        """

        selectedDetector = self._detectorComboBox.currentText()

        message = (
            f"You just changed the detector to <b><i>{selectedDetector}</i></b>.<br><br>"
            "Please ensure that the <b>config file</b> in the <i>pyFAI config</i> section is correct.<br>"
            "If necessary, load the appropriate <b>JSON</b> or <b>PONI</b> file."
        )

        qt.QMessageBox.information(self, "Configuration file verification", message)

    def getMonitorName(self) -> str | None:
        if self._monitorNameComboBox.currentText() == "-":
            return None
        return self._monitorNameComboBox.currentText()

    def setMonitorName(self, value: str | None):
        if value is None:
            self._monitorNameComboBox.setCurrentText("-")
        else:
            self._monitorNameComboBox.setCurrentText(value)

    def getSigmaClippingThreshold(self) -> float | None:
        """
        Return threshold if sigma clipping is enabled, else None.
        """
        if self._sigmaClipCheckBox.isChecked():
            return self._sigmaClipThresholdSpinBox.value()
        return None

    def _sigmaClipCheckBoxToggled(self, toggled: bool) -> None:
        """
        Update integration methods combo box according to sigma clipping.

        - Change selected method if not supported by sigma clipping
        - Enable/Disable pixel splitting methods depending on sigma clipping
        """
        noCsrPrefix = "no_csr"

        if toggled and not self._integrationMethodComboBox.currentText().startswith(
            noCsrPrefix
        ):
            self._integrationMethodComboBox.setCurrentText("no_csr_ocl_gpu")

        model = self._integrationMethodComboBox.model()
        for row in range(self._integrationMethodComboBox.count()):
            item = model.item(row)
            if not item.text().startswith(noCsrPrefix):
                flags = item.flags()
                if toggled:
                    item.setFlags(flags & ~qt.Qt.ItemIsEnabled)
                    item.setToolTip("Not available with sigma clipping")
                else:
                    item.setFlags(flags | qt.Qt.ItemIsEnabled)
                    item.setToolTip("")

    def getIntegrationMethod(self) -> str:
        return self._integrationMethodComboBox.currentText()

    def setIntegrationMethod(self, value: str):
        self._integrationMethodComboBox.setCurrentText(value)

    def getSelectedUnits(self) -> list[str]:
        """
        Return a list of selected radial units: 'q_A^-1', 'q_nm^-1' and/or '2th_deg'.
        """
        units = list()
        if self._qCheckBox.isChecked():
            unit = self._qUnitsComboBox.currentText()
            if unit == "Å⁻¹":
                units.append("q_A^-1")
            elif unit == "nm⁻¹":
                units.append("q_nm^-1")
        if self._2thCheckBox.isChecked():
            units.append("2th_deg")
        return units

    def get2ThCheckBox(self) -> qt.QCheckBox:
        """
        Return the 2theta checkbox (for test purposes).
        """
        return self._2thCheckBox

    def loadSettings(self) -> None:
        settings = qt.QSettings()

        detectorName = settings.value("options/detectorName", None)
        if detectorName in DETECTOR_NAMES:
            self._detectorComboBox.setCurrentText(detectorName)

        monitorName = settings.value("options/monitorName", None)
        if monitorName == "-" or monitorName in MONITOR_NAMES:
            self._monitorNameComboBox.setCurrentText(monitorName)

        self._sigmaClipCheckBox.setChecked(
            settings.value("options/sigmaClipping/enabled", False, type=bool)
        )
        self._sigmaClipThresholdSpinBox.setValue(
            settings.value("options/sigmaClipping/threshold", 3, type=float)
        )

        integrationMethod = settings.value("options/integrationMethod", None)
        if integrationMethod in INTEGRATION_METHODS:
            self._integrationMethodComboBox.setCurrentText(integrationMethod)

        self._qCheckBox.setChecked(settings.value("options/units/q", False, type=bool))
        selectedQUnit = settings.value("options/units/q_unit", "Å⁻¹")
        if selectedQUnit in ["Å⁻¹", "nm⁻¹"]:
            self._qUnitsComboBox.setCurrentText(selectedQUnit)

        self._2thCheckBox.setChecked(
            settings.value("options/units/2th", False, type=bool)
        )

    def saveSettings(self) -> None:
        settings = qt.QSettings()

        settings.setValue("options/detectorName", self._detectorComboBox.currentText())

        settings.setValue(
            "options/monitorName", self._monitorNameComboBox.currentText()
        )

        settings.setValue(
            "options/sigmaClipping/enabled", self._sigmaClipCheckBox.isChecked()
        )
        settings.setValue(
            "options/sigmaClipping/threshold",
            float(self._sigmaClipThresholdSpinBox.value()),
        )

        settings.setValue("options/units/q", self._qCheckBox.isChecked())
        settings.setValue("options/units/q_unit", self._qUnitsComboBox.currentText())
        settings.setValue("options/units/2th", self._2thCheckBox.isChecked())



import json
from pathlib import Path

import numpy
import pytest
import scipy.io
import logging

from ..app.integrate.mainwindow import MainWindow
from ..app.integrate.utils import ScanEntry
from .conftest import DETECTOR_SHAPE


PONI = {
    "application": "pyfai-integrate",
    "version": 3,
    "wavelength": 1.6531226457760035e-11,
    "dist": 1.014612139238891,
    "poni1": 0.21340662994315895,
    "poni2": 0.13912764384981186,
    "rot1": -0.004822100080733148,
    "rot2": 0.0005542810978441514,
    "rot3": 0.0,
    "detector": "Pilatus_CdTe_2M",
}


def _create_flats(flat_dir: Path, n=5):
    energy = numpy.linspace(0, 10, n)
    flat = numpy.ones((*DETECTOR_SHAPE, n))
    scipy.io.savemat(flat_dir / "flats.mat", mdict={"E": energy, "F": flat})
    scipy.io.savemat(flat_dir / "flats_old.mat", mdict={"Eold": energy, "Fold": flat})


def test_integrate(qtbot, tmp_path, tomo_scans, caplog):
    window = MainWindow()
    qtbot.addWidget(window)

    with open(tmp_path / "config.json", "w") as cfg:
        json.dump(PONI, cfg)
    window.setConfigFilePath(str(tmp_path / "config.json"))

    _create_flats(tmp_path)
    window.setFlatFieldDirName(tmp_path)

    window._processingGroupBox.get2ThCheckBox().click()

    output_dir = tmp_path / "PROCESSED_DATA"
    output_dir.mkdir()
    window._outputWidget.setOutputDirName(output_dir)

    # Launch processing manually (could be done via the GUI later)
    with qtbot.waitSignal(window._executor.finished, timeout=20000):
        window._processScans([ScanEntry(str(tomo_scans), 2)], local=True)

    # Overwrite the files just created

    with caplog.at_level(logging.INFO, logger="ewokscore"):
        with qtbot.waitSignal(window._executor.finished, timeout=20000):
            window._processScans([ScanEntry(str(tomo_scans), 2)], local=True)
        assert (
            "Unable to synchronously create group (name already exists)"
            not in caplog.text
        )

    integration_output_file = output_dir / f"{tomo_scans.stem}.h5"
    assert integration_output_file.exists()


def test_detector_error(qtbot, tmp_path, tomo_scans):
    window = MainWindow()
    qtbot.addWidget(window)

    with open(tmp_path / "config.json", "w") as cfg:
        json.dump(PONI, cfg)
    window.setConfigFilePath(str(tmp_path / "config.json"))

    _create_flats(tmp_path)
    window.setFlatFieldDirName(tmp_path)

    window._processingGroupBox.get2ThCheckBox().click()

    output_dir = tmp_path / "PROCESSED_DATA"
    output_dir.mkdir()
    window._outputWidget.setOutputDirName(output_dir)

    # Launch processing manually (could be done via the GUI later)
    with qtbot.waitSignals(
        [window._executor.jobSubmitted, window._executor.finished],
        order="strict",
        timeout=20000,
    ) as blocker:
        window._processScans([ScanEntry(str(tomo_scans), 3)], local=True)

    # See https://pytest-qt.readthedocs.io/en/latest/signals.html#getting-emitted-signals-and-arguments
    job_item = blocker.all_signals_and_args[0].args[0]
    with pytest.raises(
        RuntimeError,
        match="Execution failed for ewoks task 'integrate' .*: '/3.1/instrument/p3/data'",
    ):
        job_item.getFuture().result()
